#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
#
# Copyright (c) 2019, Intel Corporation. All rights reserved.
#
# Author: Janusz Jankowski <janusz.jankowski@linux.intel.com>

# Loads current config file and updates with values from another
# config file. Configs that are not set in neither of files are set
# to default value.
#
# Usage:
#   overrideconfig.py KCONFIG_FILENAME CONFIG_OVERRIDE_FILE

import os
import sys

import kconfiglib


def main():
    if len(sys.argv) < 3:
        sys.exit("usage: {} KCONFIG_FILENAME CONFIG_OVERRIDE_FILE"
	    .format(sys.argv[0]))

    kconf = kconfiglib.Kconfig(sys.argv[1])
    kconf.load_config()

    kconf.disable_override_warnings()
    kconf.disable_redun_warnings()

    kconf.load_config(filename=sys.argv[2], replace=False)

    kconf.enable_override_warnings()
    kconf.enable_redun_warnings()

    kconf.write_config()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause

# Add rounded new file size to accumulated module size cache

import argparse
import pathlib
import os
from elftools.elf.elffile import ELFFile
from elftools.elf.constants import SH_FLAGS

args = None

def parse_args():
	global args

	parser = argparse.ArgumentParser(description='Add a file size to a sum in a file')

	parser.add_argument("-i", "--input", required=True, type=str,
						help='Object file name')
	parser.add_argument("-s", "--size-file", required=True, type=str,
						help='File to store accumulated size')

	args = parser.parse_args()

def get_elf_size(elf_name):
	start = 0xffffffff
	# SOF_MODULE_DRAM_LINK_END
	min_start = 0x08000000
	end = 0
	with open(elf_name, 'rb') as f_elf:
		elf = ELFFile(f_elf)

		for section in elf.iter_sections():
			s_flags = section.header['sh_flags']

			if not s_flags & SH_FLAGS.SHF_ALLOC:
				continue

			# Ignore detached sections, to be used in DRAM, their addresses
			# are below min_start
			if section.header['sh_addr'] < min_start:
				continue

			if section.header['sh_addr'] < start:
				start = section.header['sh_addr']
			if section.header['sh_addr'] + section.header['sh_size'] > end:
				end = section.header['sh_addr'] + section.header['sh_size']

		size = end - start

	return size

def main():
	global args

	parse_args()

	f_output = pathlib.Path(args.size_file)

	try:
		with open(f_output, 'r') as f_size:
			size = int(f_size.read().strip(), base = 0)
	except OSError:
		size = 0

	size += get_elf_size(args.input) + 0xfff
	# align to a page border
	size &= ~0xfff

	# Failure will raise an exception
	f_size = open(f_output, "w")
	f_size.write(f'0x{size:x}\n')

if __name__ == "__main__":
	main()



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
#
# Copyright (c) 2019, Intel Corporation. All rights reserved.
#
# Author: Michal Jerzy Wierzbicki <michalx.wierzbicki@linux.intel.com>
#         Adrian Bonislawski <adrian.bonislawski@linux.intel.com>

# Tool for processing FW stack dumps.
# For more detailed useage, use --help option.

from __future__  import print_function
import argparse
import struct
import os
import sys
import itertools
import re
import shutil
import time
from ctypes      import LittleEndianStructure, BigEndianStructure, c_uint32, c_char
from collections import namedtuple
from operator    import attrgetter
from functools   import partial

def stderr_print(*args, **kwargs):
	print(*args, file=sys.stderr, **kwargs)

def stdout_print(*args, **kwargs):
	print(*args, file=sys.stdout, **kwargs)

try:
	from sty import fg, bg, ef, rs, Rule, Render
	CAN_COLOUR=True
except:
	CAN_COLOUR=False

ArchDefinition = namedtuple('ArchDefinition', ['name', 'bitness', 'endianness'])
VALID_ARCHS = {}
[VALID_ARCHS.update({archdef.name : archdef})
	for archdef in [ArchDefinition(*tup) for tup in
	[
		( 'LE32bit',
			32,
			LittleEndianStructure, ),
		( 'LE64bit',
			64,
			LittleEndianStructure, ),
#		( 'BE32bit', #untested, treat as not implemented
#			32,
#			BigEndianStructure, ),
#		( 'BE64bit', #untested, treat as not implemented
#			64,
#			BigEndianStructure, ),
	]]
]

# Exception causes:
# CODE: [Exception cause, excvaddr loaded]
EXCCAUSE_CODE = {
	0: ["IllegalInstructionCause: Illegal instruction", False],
	1: ["SyscallCause: SYSCALL instruction", True],
	2: ["InstructionFetchErrorCause: Processor internal physical address or data error during instruction fetch", True],
	3: ["LoadStoreErrorCause: Processor internal physical address or data error during load or store", True],
	4: ["Level1InterruptCause: Level-1 interrupt as indicated by set level-1 bits in the INTERRUPT register", False],
	5: ["AllocaCause: MOVSP instruction, if caller's registers are not in the register file", False],
	6: ["IntegerDivideByZeroCause: QUOS, QUOU, REMS, or REMU divisor operand is zero", False],
	8: ["PrivilegedCause: Attempt to execute a privileged operation when CRING ? 0", False],
	9: ["LoadStoreAlignmentCause: Load or store to an unaligned address", True],
	12: ["InstrPIFDataErrorCause: PIF data error during instruction fetch", True],
	13: ["LoadStorePIFDataErrorCause: ynchronous PIF data error during LoadStore access", True],
	14: ["InstrPIFAddrErrorCause: PIF address error during instruction fetch", True],
	15: ["LoadStorePIFAddrErrorCause: Synchronous PIF address error during LoadStore access", True],
	16: ["InstTLBMissCause: Error during Instruction TLB refill", True],
	17: ["InstTLBMultiHitCause: Multiple instruction TLB entries matched", True],
	18: ["InstFetchPrivilegeCause: An instruction fetch referenced a virtual address at a ring level less than CRING", True],
	20: ["InstFetchProhibitedCause: An instruction fetch referenced a page mapped with an attribute that does not permit instruction fetch", True],
	24: ["LoadStoreTLBMissCause: Error during TLB refill for a load or store", True],
	25: ["LoadStoreTLBMultiHitCause: Multiple TLB entries matched for a load or store", True],
	26: ["LoadStorePrivilegeCause: A load or store referenced a virtual address at a ring level less than CRING", True],
	28: ["LoadProhibitedCause: A load referenced a page mapped with an attribute that does not permit loads", True],
	29: ["StoreProhibitedCause: A store referenced a page mapped with an attribute that does not permit stores", True],
	32: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False],
	33: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False],
	34: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False],
	35: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False],
	36: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False],
	37: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False],
	38: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False],
	39: ["CoprocessornDisabled: Coprocessor n instruction when cpn disabled. n varies 0..7 as the cause varies 32..39", False]
}

def valid_archs_print():
	archs = ''.join("{0}, ".format(x) for x in VALID_ARCHS)
	archs = archs[:len(archs)-2]
	return "{0}.".format(archs)

TERM_SIZE = shutil.get_terminal_size((120, 20))
AR_WINDOW_WIDTH = 4
IS_COLOUR=False

class argparse_readable_file( argparse.Action):
	def raise_error(self, filepath, reason):
		raise argparse.ArgumentTypeError(
			"is_readable_file:{0} {1}".format(
				filepath,
				reason
			))
	def is_readable_file(self, filepath):
		if not os.path.isfile(filepath):
			self.raise_error(filepath, "is not a valid path")
		if os.access(filepath, os.R_OK):
			return True
		else:
			self.raise_error(filepath, "is not a readable file")
		return False
	def __call__(self, parser, namespace, values, option_string=None):
		filepath = values[0]
		if (self.is_readable_file(filepath)):
			setattr(namespace, self.dest, filepath)

class argparse_writeable_file(argparse.Action):
	def raise_error(self, filepath, reason):
		raise argparse.ArgumentTypeError(
			"is_writeable_file:{0} {1}".format(
				filepath,
				reason
			))
	def is_writeable_file(self, filepath):
		absdir = os.path.abspath(os.path.dirname(filepath))
		if os.path.isdir(absdir) and not os.path.exists(filepath):
			return True
		else:
			if not os.path.isfile(filepath):
				self.raise_error(filepath, "is not a valid path")
			if os.access(filepath, os.W_OK):
				return True
			else:
				self.raise_error(filepath, "is not a writeable file")
		return False
	def __call__(self, parser, namespace, values, option_string=None):
		filepath = values[0]
		if (self.is_writeable_file(filepath)):
			setattr(namespace, self.dest, filepath)
		else:
			self.raise_error(
				filepath,
				"failed to determine whether file is writeable"
			)

class argparse_architecture(  argparse.Action):
	def raise_error(self, value, reason):
		raise argparse.ArgumentTypeError(
			"architecture: {0} {1}".format(
				value,
				reason
			))
	def is_valid_architecture(self, value):
		if value in VALID_ARCHS:
			return True
		return False
	def __call__(self, parser, namespace, values, option_string=None):
		value = values[0]
		if (self.is_valid_architecture(value)):
			setattr(namespace, self.dest, VALID_ARCHS[value])
		else:
			self.raise_error(
				value,
				"is invalid architecture. Valid architectures are: {0}".format(valid_archs_print())
			)

def parse_params():
	parser = argparse.ArgumentParser(
		description="Tool for processing FW stack dumps."
			+" In verbose mode it prints DSP registers, and function call"
			+" addresses in stack up to that which caused core dump."
			+" It then prints either to file or to stdin all gdb"
			+" commands unwrapping those function call addresses to"
			+" function calls in  human readable format."
	)
	ArgTuple = namedtuple('ArgTuple', ['name', 'optionals', 'parent'])
	ArgTuple.__new__.__defaults__ = ((), {}, parser)

	# below cannot be empty once declared
	outputMethod = parser.add_mutually_exclusive_group(required=True)
	inputMethod  = parser.add_mutually_exclusive_group()
	[parent.add_argument(*name, **optionals)
		for name, optionals, parent in sorted([ArgTuple(*x) for x in
			[
				( ( '-a', '--arch'       , ), {
						'type'   : str,
						'help'   :'determine architecture of dump file; valid archs are:  {0}'
							.format(valid_archs_print()),
						'action' : argparse_architecture,
						'nargs'  : 1,
						'default': VALID_ARCHS['LE64bit'],
					},),
				( ( '-v', '--verbose'    , ), {
						'help'  :'increase output verbosity',
						'action':'store_true',
					},),
				( (       '--stdin'      , ), {
						'help'   :'input is from stdin',
						'action' : 'store_true',
					},
					inputMethod),
				( ( '-o', '--outfile'    , ), {
						'type'   : str,
						'help'   :'output is to FILE',
						'action' : argparse_writeable_file,
						'nargs'  : 1,
					},
					outputMethod),
				( (       '--stdout'     , ), {
						'help'   :'output is to stdout',
						'action' :'store_true',
					},
					outputMethod),
				( ( '-c', '--colour', ), {
						'help'  :'set output to be colourful!',
						'action':'store_true',
					},),
				( ( '-l', '--columncount', ), {
						'type'  : int,
						'help'  :'set how many colums to group the output in, ignored without -v',
						'action':'store',
						'nargs' : 1,
					},),
				( ( '-i', '--infile'     , ), {
						'type'   : str,
						'help'   :'path to sys dump bin',
						'action' : argparse_readable_file,
						'nargs'  : 1,
					},
					inputMethod),
			]],
			key=lambda argtup: (argtup.parent.__hash__(), argtup.name)
		)
	]

	parsed = parser.parse_args()

	if parsed.columncount and not parsed.verbose:
		stderr_print("INFO: -l option will be ignored without -v")

	return parsed

def chunks(l, n):
	return [l[i:i + n] for i in range(0, len(l), n)]

def flaten(l):
	return [item for sublist in l for item in sublist]

def raiseIfArchNotValid(arch):
	if arch not in VALID_ARCHS.values():
		raise ValueError(
			"CoreDumpFactory: {0} not in valid architectures: {1}"
				.format(arch, valid_archs_print())
		)
	endiannesses = [arch.endianness for arch in VALID_ARCHS.values()]
	if arch.endianness not in endiannesses:
		raise ValueError(
			"CoreDumpFactory: {0} not in valid endiannesses: {1}"
				.format(endianness, endiannesses)
		)

def FileInfoFactory(arch, filename_length):
	raiseIfArchNotValid(arch)
	class FileInfo(arch.endianness):
		_fields_ = [
			("hdr",  c_uint32),
			("code",  c_uint32),
			("filename", filename_length * c_char),
			("line_no",  c_uint32)
		]
		def __str__(self):
			return "{}:{:d}".format(self.filename.decode(), self.line_no)
	if FileInfo is None:
		raise RuntimeError(
			"FileInfoFactory: failed to produce FileInfo({0})"
				.format(arch.name)
		)
	return FileInfo

class Colourer():
	#TODO: Add detection of 8bit/24bit terminal
	#      Add 8bit/24bit colours (with flag, maybe --colour=24bit)
	#      Use this below as fallback only
	__print = partial(stdout_print)
	if CAN_COLOUR == True:
		__style = {
			'o' : fg.red,
			'O' : fg.yellow,
			'y' : fg.blue,
			'Y' : fg.cyan,
			'D' : fg.white + bg.red,
		}
	matchings          = []
	matchingsInParenth = []
	matchingsInStderr  = []
	def __init__(self):
		if CAN_COLOUR == True:
			self.matchings = [
				(
					lambda x: self.enstyleNumHex(x.group()),
					re.compile(r'\b([A-Fa-f0-9]{8})\b')
				),
				(
					lambda x: '0x' + self.enstyleNumHex(x.group(2)),
					re.compile(r'(0x)([A-Fa-f0-9]{1,})')
				),
				(
					lambda x: self.enstyleNumBin(x.group()),
					re.compile(r'\b(b[01]+)\b')
				),
				(
					r'\1' +
					r'\2' +
					self.enstyle(           fg.green , r'\3') ,
					re.compile(r'(\#)(ar)([0-9]+)\b')
				),
				(
					self.enstyle(bg.green            , r'\1') +
					rs.all + '\n',
					re.compile(r'(\(xt-gdb\)\ *)')
				),
				(
					r'\1' +
					r'\2' +
					self.enstyle(           fg.green   , r'\3') +
					r':'  +
					self.enstyle(           fg.magenta , r'\4') ,
					re.compile(r'(\bat\b\ *)(.+/)?(.*):([0-9]+)')
				),
				(
					lambda x:\
						'in '+
						self.enstyle(fg.green,  x.group(2))+
						self.enstyleFuncParenth(x.group(3)),
					re.compile(r'(\bin\b\ *)([^\ ]+)\ *(\(.*\))')
				),
			]
			self.matchingsInParenth = [
				(
					self.enstyle(           fg.yellow   , r'\1') +
					self.enstyle(           fg.magenta  , r'=' ) +
					r'\2',
					re.compile(r'([\-_a-zA-Z0-9]+)\ *=\ *([^,]+)')
				),
				(
					self.enstyle(           fg.magenta  , r'\1') ,
					re.compile(r'(\ *[\(\)]\ *)')
				),
				(
					self.enstyle(           fg.magenta  , r', ') ,
					re.compile(r'(\ *,\ *)')
				),
			]
			self.matchingsInStderr = [
				(
					self.enstyle(bg.yellow  + fg.black    , r'\1') ,
					re.compile(r'([Ww]arning)')
				),
				(
					self.enstyle(bg.red     + fg.black    , r'\1') ,
					re.compile(r'([Ee]rror)')
				),
				(
					self.enstyle(bg.magenta + fg.black    , r'\1') ,
					re.compile(r'([Ff]atal)')
				),
			]

	def toGroup(self, txt):
		return [ (label, sum(1 for _ in group))
				for label, group in itertools.groupby(txt) ]

	def leadingZero(self, txt, char):
		result = ""
		groups = self.toGroup(txt)
		lead = 0
		if groups[0][0] == '0':
			lead = min(4, groups[0][1])
			result += char.lower() *    lead
		result +=     char.upper() * (4-lead)
		return result

	def findSub(self, txt, mask, sub, char):
		pos = txt.find(sub)
		if pos >= 0:
			return mask[:pos] + char * len(sub) + mask[(len(sub)+pos):]
		else:
			return mask

	def enstyleFuncParenth(self, txt):
		result = txt
		for repl, regex in self.matchingsInParenth:
			result = re.sub(regex, repl, result)
		return result 
		
	def enstyleNumBin(self, txt):
		result = rs.all + bg.magenta + "b"
		prev = ""
		for c in txt[1:]:
			if prev != c:
				prev = c
				result += rs.all
				if c == "0":
					result += fg.red
			result += c
		result += rs.all
		return result

	def enstyleNumHex(self, txt):
		if len(txt) < 8:
			txt = (8-len(txt))*'0' + txt
		p1 = 'o'
		p2 = 'y'
		if txt == "00000000":
			styleMask = p1 * 8
		elif txt.lower() == "deadbeef":
			styleMask = "DDDDDDDD"
		else:
			styleMask = "".join(
				[self.leadingZero(string, style)
				for string, style in [
					(txt[:4], p1),
					(txt[4:], p2),
			]])
			styleMask = "".join(
				[self.findSub(txt, styleMask, string, style)
				for string, style in [
					('dead', 'D'),
			]])

		result = ""
		thisstyle = ''
		for iter, style in enumerate(styleMask):
			if thisstyle != style:
				thisstyle = style
				result += rs.all + self.__style[thisstyle]
			result += txt[iter]
		result += rs.all
		return result

	def enstyleStderr(self, txt):
		if txt is None:
			return ''
		result = txt
		for repl, regex in self.matchingsInStderr:
			result = re.sub(regex, repl, result)
		for repl, regex in self.matchings:
			result = re.sub(regex, repl, result)
		return fg.red + result + rs.all

	def enstyle(self, style, txt):
		return style + txt + rs.all

	def produce_string(self, txt):
		result = txt
		for repl, regex in self.matchings:
			result = re.sub(regex, repl, result)
		return result

	def print(self, txt):
		self.__print(self.produce_string(txt))

def CoreDumpFactory(dsp_arch):
	raiseIfArchNotValid(dsp_arch)
	class CoreDump(dsp_arch.endianness):
		_fields_ = [(x, c_uint32) for x in
				[
					# struct sof_ipc_dsp_oops_arch_hdr {
					"arch",
					"totalsize",
					# }
					# struct sof_ipc_dsp_oops_plat_hdr {
					"configidhi",
					"configidlo",
					"numaregs",
					"stackoffset",
					"stackptr",
					# }
					"exccause",
					"excvaddr",
					"ps"
				]
				+ ["epc" + str(x) for x in range(1,7+1)]
				+ ["eps" + str(x) for x in range(2,7+1)]
				+ [
					"depc",
					"intenable",
					"interrupt",
					"sar",
					"debugcause",
					"windowbase",
					"windowstart",
					"excsave1" # to
				]
			] + [
				("a", dsp_arch.bitness * c_uint32)
			]

		def __init__(self, columncount):
			self.dsp_arch = dsp_arch
			self._fields_
			self.ar_regex = re.compile(r'ar[0-9]+')
			# below: smart column count
			self._longest_field = len(max([x[0] for x in self._fields_], key=len))
			if columncount is not None:
				self.columncount = max (1, int(columncount[0]))
			else:
				self.columncount = max(1,
					int(TERM_SIZE[0]/(self._longest_field + 2 + 2 * AR_WINDOW_WIDTH + 2))
				)
			self.columncount_ar = (
				self.columncount
					if self.columncount <= AR_WINDOW_WIDTH else
				AR_WINDOW_WIDTH * int(self.columncount/AR_WINDOW_WIDTH)
			)

		def __windowbase_shift(self, iter, direction):
			return (iter + self.windowbase * AR_WINDOW_WIDTH * direction)\
				% self.dsp_arch.bitness
		def windowbase_shift_left(self, iter):
			return self.__windowbase_shift(iter, -1)
		def windowbase_shift_right(self, iter):
			return self.__windowbase_shift(iter,  1)

		def reg_from_string(self, string):
			if self.ar_regex.fullmatch(string):
				return self.a[self.windowbase_shift_left(int(string[2:]))]
			else:
				return self.__getattribute__(string)

		def __str__(self):
			string = ""
			string += "exccause"
			return string

		def to_string(self, is_gdb):
			# in case windowbase in dump has invalid value
			windowbase_shift = min(
				self.windowbase * AR_WINDOW_WIDTH,
				self.dsp_arch.bitness
			)
			# flatten + chunk enable to smartly print in N columns
			string = ''.join([self.fmt(is_gdb, x)
				for x in flaten(
					[chunks(word, self.columncount) for word in [
						["arch", "totalsize", "stackptr", "stackoffset"],
						["configidhi", "configidlo", "numaregs"],
					]]
				)
			])

			string += "\n# CPU registers:\n\n"

			string += ''.join([self.fmt(is_gdb, x)
				for x in flaten(
					[chunks(word, self.columncount) for word in [
						["exccause", "excvaddr", "ps"],
						["epc" + str(x) for x in range(1,7+1)],
						["eps" + str(x) for x in range(2,7+1)],
						["depc", "intenable", "interrupt", "sar", "debugcause"],
						["windowbase", "windowstart"],
						["excsave1"],
					]] +
					[chunks(word, self.columncount_ar) for word in [
						["ar" + str(x) for x in itertools.chain(
							range(   windowbase_shift, self.dsp_arch.bitness),
							range(0, windowbase_shift),
						)]
					]]
				)
			])

			if not is_gdb:
				string += "\n"
			return string

		def fmt_gdb_command(self):
			return "set ${}=0x{:08x}\n"

		def fmt_pretty_form(self, separator = "|"):
			return separator + "{:" + str(self._longest_field) + "} {:08x} "

		def fmt_separator(self, name):
			separator = "# "
			return separator

		def fmt_pretty_auto(self, name):
			return self.fmt_pretty_form(self.fmt_separator(name))

		def fmt(self, is_gdb, names):
			if is_gdb:
				fmtr = lambda name: self.fmt_gdb_command()
			else:
				fmtr = lambda name: self.fmt_pretty_auto(name)

			string = ""
			for name in names:
				string += fmtr(name).format(
					name, self.reg_from_string(name)
				)
			if not is_gdb:
				string += "\n"
			return string

		def windowstart_process(self):
			string = ""
			binary = "{0:b}".format(self.windowstart)
			bit_start = len(binary)-1-self.windowbase
			fnc_num = 0
			header = ""

			for it, c in enumerate(binary[bit_start:]):
				if c != "0":
					header += str(fnc_num)
					fnc_num += 1
				else:
					header += " "
			for it, c in enumerate(binary[:bit_start]):
				if c != "0":
					header += str(fnc_num)
					fnc_num += 1
				else:
					header += " "
			header = header[self.windowbase+1:]+ header[:self.windowbase+1]

			string += "# windowbase: {:0X}\n".format(self.windowbase)
			string += "#               {0}\n".format(header)
			string += "# windowstart: b{0}\n\n".format(binary)
			string += "#      reg         a0         a1\n"
			string += "#                  (return)   (sptr)\n"
			string += "#      ---         --------   -------\n"
			fnc_num = 0
			for iter, digit in enumerate(binary[bit_start:]):
				if (digit == '1'):
					reg = "ar{0}".format(AR_WINDOW_WIDTH * (self.windowbase - iter))
					reg1 = "ar{0}".format(AR_WINDOW_WIDTH * (self.windowbase - iter) + 1)
					string += "# {0:2d} ".format(++fnc_num)
					string += self.fmt_pretty_auto(reg).format(
						reg, self.reg_from_string(reg)
					) + "  {:0x} ".format(self.reg_from_string(reg1)) + "\n"
					fnc_num += 1
			for iter, digit in enumerate(binary[:bit_start]):
				if (digit == '1'):
					reg = "ar{0}".format(AR_WINDOW_WIDTH * (len(binary) - 1 - iter))
					reg1 = "ar{0}".format(AR_WINDOW_WIDTH * (len(binary) - 1 - iter) + 1)
					string += "# {0:2d} ".format(++fnc_num)
					string += self.fmt_pretty_auto(reg).format(
						reg, self.reg_from_string(reg)
					) + "  {:0x} ".format(self.reg_from_string(reg1)) + "\n"
					fnc_num += 1
			return string

	if CoreDump is None:
		raise RuntimeError(
			"CoreDumpFactory: failed to produce CoreDump({0})"
				.format(dsp_arch.name)
		)
	return CoreDump

class CoreDumpReader(object):
	def __init__(self, args):
		self.core_dump    = CoreDumpFactory(args.arch)(
			args.columncount
		)
		self.file_info    = FileInfoFactory(args.arch, 32)()

		if IS_COLOUR:
			colourer = Colourer()
		if args.verbose:
			verbosePrint =\
					colourer.print\
				if IS_COLOUR else\
					stdout_print
		else:
			verbosePrint = lambda *discard_this: None

		if   args.stdout:
			stdoutOpen  = lambda: None
			stdoutPrint = print
			stdoutClose = lambda: None
		elif args.outfile:
			#TODO: open file in stdOutOpen
			stdoutDest  = open(args.outfile, "w")
			stdoutOpen  = lambda: None
			stdoutPrint = stdoutDest.write
			stdoutClose = stdoutDest.close
		else:
			raise RuntimeError("CoreDumpReader: No output method.") 

		if   args.stdin:
			inStream = lambda: sys.stdin.buffer
		elif args.infile:
			inStream = lambda: open(args.infile, "rb")
		else:
			raise RuntimeError("CoreDumpReader: No input method.") 

		with inStream() as cd_file:
			[cd_file.readinto(x) for x in [
				self.core_dump,
				self.file_info
			]]
			self.stack = cd_file.read()

		verbosePrint("# Core header:\n")
		verbosePrint(self.core_dump.to_string(0))

		verbosePrint(self.core_dump.windowstart_process())

		stack_base = self.core_dump.stackptr
		stack_dw_num = int(len(self.stack)/AR_WINDOW_WIDTH)
		verbosePrint("# Stack dumped from {:08x} dwords num {:d}"
			.format(stack_base, stack_dw_num))

		stdoutOpen()

		# disable confirmation request for undefined breakpoint
		stdoutPrint("set confirm off\n")

		# for XTOS SOF build
		stdoutPrint("break _MemErrorVector\n")
		# for Zephyr SOF build
		stdoutPrint("break _MemoryExceptionVector_text_start\n")

		stdoutPrint("run\n")
		stdoutPrint(self.core_dump.to_string(1))

		#TODO: make this elegant
		for dw, addr in [(
			struct.unpack("I", self.stack[i*AR_WINDOW_WIDTH : (i+1)*AR_WINDOW_WIDTH])[0],
			stack_base + i*AR_WINDOW_WIDTH
		) for i in range(0, stack_dw_num)]:
			stdoutPrint("set *0x{:08x}=0x{:08x}\n"
				.format(addr, dw))

                # Exccause 63 is reserved for panics; the other causes come
                # from actual exceptions
		if self.core_dump.exccause != 63:
			verbosePrint("\n# *EXCEPTION*\n")
			verbosePrint("# exccause: " + EXCCAUSE_CODE[self.core_dump.exccause][0])
			if EXCCAUSE_CODE[self.core_dump.exccause][1]:
				verbosePrint("# excvaddr: " + str(self.core_dump.excvaddr))
			stdoutPrint('p "Exception location:"\nlist *$epc1\n');
		else:
			verbosePrint("# Location: " + str(self.file_info));

		# TODO: if excsave1 is not empty, pc should be set to that value
		# (exception mode, not forced panic mode)
		stdoutPrint('set $pc=&arch_dump_regs_a\np "backtrace"\nbacktrace\n')
		stdoutClose()

if __name__ == "__main__":
	args = parse_params()
	if args.colour:
		if CAN_COLOUR:
			IS_COLOUR=True
		else:
			stderr_print("INFO: Cannot color the output: module 'sty' not found")
	CoreDumpReader(args)




#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Creates new module based on template module.

import os
import sys
import shutil
import uuid
import re

def process_directory(directory_path, old_text, new_text):
    """
    Recursively walks through a directory to rename files and replace content.
    This function processes files first, then directories, to allow for renaming
    of the directories themselves after their contents are handled.
    """
    # Walk through the directory tree
    for dirpath, dirnames, filenames in os.walk(directory_path, topdown=False):
        # --- Process Files ---
        for filename in filenames:
            original_filepath = os.path.join(dirpath, filename)

            # a) Replace content within files
            # Only process C/H files for content replacement
            if filename.endswith(('.c', '.h', '.cpp', '.hpp', '.txt', '.toml', 'Kconfig')):
                replace_text_in_file(original_filepath, old_text, new_text)
            if filename.endswith(('.c', '.h', '.cpp', '.hpp', '.txt', '.toml', 'Kconfig')):
                replace_text_in_file(original_filepath, old_text.upper(), new_text.upper())
            if filename.endswith(('.c', '.h', '.cpp', '.hpp', '.txt', '.toml', 'Kconfig')):
                replace_text_in_file(original_filepath, "template", new_text)
            if filename.endswith(('.c', '.h', '.cpp', '.hpp', '.txt', '.toml', 'Kconfig')):
                replace_text_in_file(original_filepath, "TEMPLATE", new_text.upper())

            # b) Rename the file if its name contains the template text
            if "template" in filename:
                new_filename = filename.replace("template", new_text)
                new_filepath = os.path.join(dirpath, new_filename)
                print(f"  -> Renaming file: '{original_filepath}' to '{new_filepath}'")
                os.rename(original_filepath, new_filepath)

def replace_text_in_file(filepath, old_text, new_text):
    """
    Replaces all occurrences of old_text with new_text in a given file.
    """
    try:
        # Read the file content
        with open(filepath, 'r', encoding='utf-8') as file:
            content = file.read()

        # Perform the replacement
        new_content = content.replace(old_text, new_text)

        # If content has changed, write it back
        if new_content != content:
            print(f"  -> Updating content in: '{filepath}'")
            with open(filepath, 'w', encoding='utf-8') as file:
                file.write(new_content)

    except Exception as e:
        print(f"  -> [WARNING] Could not process file '{filepath}'. Reason: {e}")

def insert_uuid_name(filepath: str, new_name: str):
    """
    Inserts a newly generated UUID and a given name into a file, maintaining
    alphabetical order by name. Ignores and preserves lines starting with '#'.

    Args:
        filepath (str): The path to the file to be updated.
        new_name (str): The name to associate with the new UUID.
    """
    data_entries = []
    other_lines = []  # To store comments and blank lines

    # --- 1. Read existing entries and comments from the file ---
    try:
        with open(filepath, 'r', encoding='utf-8') as f:
            for line in f:
                stripped_line = line.strip()
                # Check for comments or blank lines
                if not stripped_line or stripped_line.startswith('#'):
                    other_lines.append(line)  # Preserve the original line with newline
                    continue

                # It's a data line, so process it
                # Split only on the first space to handle names that might contain spaces
                parts = stripped_line.split(' ', 1)
                if len(parts) == 2:
                    data_entries.append((parts[0], parts[1]))
    except FileNotFoundError:
        print(f"File '{filepath}' not found. A new file will be created.")
    except Exception as e:
        print(f"An error occurred while reading the file: {e}")
        return

    # --- 2. Check if the name already exists in data entries ---
    if any(name == new_name for _, name in data_entries):
        print(f"Name '{new_name}' already exists in the file. No changes made.")
        return

    # --- 3. Add the new entry to the data list ---
    new_uuid = str(uuid.uuid4())

    # We split the string from the right at the last hyphen and then join the parts.
    parts = new_uuid.rsplit('-', 1)
    custom_format_uuid = ''.join(parts)

    data_entries.append((custom_format_uuid, new_name))
    print(f"Generated new entry: {new_uuid} {new_name}")

    # --- 4. Sort the list of data entries by name (the second element of the tuple) ---
    data_entries.sort(key=lambda item: item[1])

    # --- 5. Write the comments and then the sorted data back to the file ---
    try:
        with open(filepath, 'w', encoding='utf-8') as f:
            # Write all the comments and other non-data lines first
            for line in other_lines:
                f.write(line)

            # Write the sorted data entries
            for entry_uuid, entry_name in data_entries:
                f.write(f"{entry_uuid} {entry_name}\n")
        print(f"Successfully updated '{filepath}'.")
    except Exception as e:
        print(f"An error occurred while writing to the file: {e}")

# Define the markers for the managed block in the CMakeLists.txt file.
CMAKE_START_MARKER = "	# directories and files included conditionally (alphabetical order)"
CMAKE_END_MARKER = "	# end of directories and files included conditionally (alphabetical order)"

def insert_cmake_rule(filepath: str, kconfig_option: str, subdir_name: str):
    """
    Reads a CMakeLists.txt file, adds a new rule, and writes it back.

    Args:
        filepath (str): Path to the CMakeLists.txt file.
        kconfig_option (str): The Kconfig flag to check.
        subdir_name (str): The subdirectory to add.
    """
    if not os.path.exists(filepath):
        print(f"[ERROR] File not found at: '{filepath}'")
        return

    # --- 1. Read the entire file content ---
    with open(filepath, 'r', encoding='utf-8') as f:
        lines = f.readlines()

    # --- 2. Find the start and end of the managed block ---
    try:
        start_index = lines.index(CMAKE_START_MARKER + '\n')
        end_index = lines.index(CMAKE_END_MARKER + '\n')
    except ValueError:
        print(f"[ERROR] Could not find the required marker blocks in '{filepath}'.")
        print(f"Please ensure the file contains both '{CMAKE_START_MARKER}' and '{CMAKE_END_MARKER}'.")
        return

    # --- 3. Extract the lines before, during, and after the block ---
    lines_before = lines[:start_index + 1]
    block_lines = lines[start_index + 1 : end_index]
    lines_after = lines[end_index:]

    # --- 4. Parse the existing rules within the block ---
    rules = []
    # Regex to find: if(CONFIG_NAME) ... add_subdirectory(subdir) ... endif()
    # This is robust against extra whitespace and blank lines.
    block_content = "".join(block_lines)
    pattern = re.compile(
        r"if\s*\((?P<kconfig>CONFIG_[A-Z0-9_]+)\)\s*"
        r"add_subdirectory\s*\((?P<subdir>[a-zA-Z0-9_]+)\)\s*"
        r"endif\(\)",
        re.DOTALL
    )

    for match in pattern.finditer(block_content):
        rules.append(match.groupdict())

    # --- 5. Check if the rule already exists ---
    if any(rule['kconfig'] == kconfig_option for rule in rules):
        print(f"[INFO] Rule for '{kconfig_option}' already exists. No changes made.")
        return

    # --- 6. Add the new rule and sort alphabetically ---
    rules.append({'kconfig': kconfig_option, 'subdir': subdir_name})
    rules.sort(key=lambda r: r['kconfig'])
    print(f"Adding rule for '{kconfig_option}' -> '{subdir_name}' and re-sorting.")

    # --- 7. Rebuild the block content from the sorted rules ---
    new_block_lines = []
    for i, rule in enumerate(rules):
        new_block_lines.append(f"\tif({rule['kconfig']})\n")
        new_block_lines.append(f"\t\tadd_subdirectory({rule['subdir']})\n")
        new_block_lines.append("\tendif()\n")

    # --- 8. Assemble the new file content and write it back ---
    new_content = "".join(lines_before + new_block_lines + lines_after)
    with open(filepath, 'w', encoding='utf-8') as f:
        f.write(new_content)

    print(f"Successfully updated '{filepath}'.")

# Define the markers for the managed block in the Kconfig file.
KCONFIG_START_MARKER = "# --- Kconfig Sources (alphabetical order) ---"
KCONFIG_END_MARKER = "# --- End Kconfig Sources (alphabetical order) ---"

def insert_kconfig_source(filepath: str, source_path: str):
    """
    Reads a Kconfig file, adds a new rsource rule, and writes it back.

    Args:
        filepath (str): Path to the Kconfig file.
        source_path (str): The path to the Kconfig file to be sourced.
    """
    if not os.path.exists(filepath):
        print(f"[ERROR] File not found at: '{filepath}'")
        return

    # --- 1. Read the entire file content ---
    try:
        with open(filepath, 'r', encoding='utf-8') as f:
            lines = f.readlines()
    except Exception as e:
        print(f"[ERROR] Could not read file: {e}")
        return

    # --- 2. Find the start and end of the managed block ---
    try:
        start_index = lines.index(KCONFIG_START_MARKER + '\n')
        end_index = lines.index(KCONFIG_END_MARKER + '\n')
    except ValueError:
        print(f"[ERROR] Could not find the required marker blocks in '{filepath}'.")
        print(f"Please ensure the file contains both '{KCONFIG_START_MARKER}' and '{KCONFIG_END_MARKER}'.")
        return

    # --- 3. Extract the lines before, during, and after the block ---
    lines_before = lines[:start_index + 1]
    block_lines = lines[start_index + 1 : end_index]
    lines_after = lines[end_index:]

    # --- 4. Parse the existing rsource rules within the block ---
    source_paths = []
    # Regex to find: rsource "path/to/file"
    pattern = re.compile(r'rsource\s+"(?P<path>.*?)"')

    for line in block_lines:
        match = pattern.search(line)
        if match:
            source_paths.append(match.group('path'))

    # --- 5. Check if the source path already exists ---
    if source_path in source_paths:
        print(f"[INFO] Source path '{source_path}' already exists. No changes made.")
        return

    # --- 6. Add the new path and sort alphabetically ---
    source_paths.append(source_path)
    source_paths.sort()
    print(f"Adding source '{source_path}' and re-sorting.")

    # --- 7. Rebuild the block content from the sorted paths ---
    new_block_lines = []
    for path in source_paths:
        new_block_lines.append(f'rsource "{path}"\n')

    # --- 8. Assemble the new file content and write it back ---
    new_content = "".join(lines_before + new_block_lines + lines_after)
    with open(filepath, 'w', encoding='utf-8') as f:
        f.write(new_content)

    print(f"Successfully updated '{filepath}'.")

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

# Define the markers for the managed block in the header file.
HEADER_START_MARKER = "/* Start of modules in alphabetical order */"
HEADER_END_MARKER = "/* End of modules in alphabetical order */"

def insert_c_declaration(header_path: str, name: str) -> bool:
    """
    Inserts a C function declaration alphabetically into a header file between
    start and end markers.

    The function declaration format is:
    "void sys_comp_module_${name}_interface_init(void);"

    Args:
        header_path (str): The full path to the C header file.
        name (str): The component name to be inserted into the function declaration.
        HEADER_START_MARKER (str): The string marking the beginning of the declaration block.
        HEADER_END_MARKER (str): The string marking the end of the declaration block.

    Returns:
        bool: True if the file was modified or already up-to-date, False on error.
    """
    new_declaration = f"void sys_comp_module_{name}_interface_init(void);"
    line_to_insert = new_declaration + "\n"

    # --- 1. Read existing lines from the header file ---
    try:
        with open(header_path, 'r') as f:
            lines = f.readlines()
    except FileNotFoundError:
        print(f"Error: Header file not found at '{header_path}'", file=sys.stderr)
        return False
    except IOError as e:
        print(f"Error: Could not read file '{header_path}': {e}", file=sys.stderr)
        return False

    # --- 2. Find start and end markers ---
    start_index = -1
    end_index = -1
    for i, line in enumerate(lines):
        if HEADER_START_MARKER in line:
            start_index = i
        elif HEADER_END_MARKER in line:
            end_index = i
            break  # Assume end marker always comes after the start marker

    if start_index == -1 or end_index == -1:
        print(f"Error: Start ('{HEADER_START_MARKER}') or end ('{HEADER_END_MARKER}') marker not found.", file=sys.stderr)
        return False
    if end_index <= start_index:
        print(f"Error: End marker appears before start marker.", file=sys.stderr)
        return False

    # --- 3. Check if the declaration already exists within the block ---
    # Isolate the block of lines where declarations are allowed
    declaration_block = lines[start_index + 1 : end_index]
    if line_to_insert in declaration_block:
        print(f"Declaration for '{name}' already exists within the block. No changes made.")
        return True

    # --- 4. Find the correct alphabetical insertion point within the block ---
    relative_insert_index = len(declaration_block)
    for i, line in enumerate(declaration_block):
        # Compare with the stripped content of each line in the block
        if line.strip() > new_declaration:
            relative_insert_index = i
            break

    # Calculate the final insertion index relative to the whole file
    final_insert_index = start_index + 1 + relative_insert_index

    # --- 5. Insert the new line into the main list of lines ---
    lines.insert(final_insert_index, line_to_insert)
    print(f"Inserting declaration for '{name}' at line {final_insert_index + 1}.")

    # --- 6. Write the modified list back to the file ---
    try:
        with open(header_path, 'w') as f:
            f.writelines(lines)
    except IOError as e:
        print(f"Error: Could not write to file '{header_path}': {e}", file=sys.stderr)
        return False

    return True

def main():
    """
    Main function to drive the script logic.
    """
    print("--- SOF SDK New Module Creator ---")

    # Argument Validation ---
    if len(sys.argv) != 2:
        print("\n[ERROR] Invalid number of arguments.")
        print("Usage: sdk_create_module.py <new_module_name>")
        sys.exit(1)

    # Configuration --- paths are with respect to script dir
    modules_root =  SCRIPT_DIR + "/../src/audio"
    template_name = "template"
    new_module_name = sys.argv[1]
    uuid_file = SCRIPT_DIR + "/../uuid-registry.txt"
    cmake_file = SCRIPT_DIR + "/../src/audio/CMakeLists.txt"
    kconfig_file = SCRIPT_DIR + "/../src/audio/Kconfig"
    component_file = SCRIPT_DIR + "/../src/include/sof/audio/component.h"

    template_dir = os.path.join(modules_root, template_name)
    new_module_dir = os.path.join(modules_root, new_module_name)

    print(f"\nConfiguration:")
    print(f"  - Modules Root:   '{modules_root}'")
    print(f"  - Template Name:  '{template_name}'")
    print(f"  - New Module Name:'{new_module_name}'")
    print(f"  - UUID file:      '{uuid_file}'")
    print(f"  - Cmake file:    '{cmake_file}'")
    print(f"  - Kconfig file:  '{kconfig_file}'")
    print(f"  - Component file: '{component_file}'")

    # Check for Pre-existing Directories ---
    if not os.path.isdir(template_dir):
        print(f"\n[ERROR] Template directory not found at: '{template_dir}'")
        sys.exit(1)

    if os.path.exists(new_module_dir):
        print(f"\n[ERROR] A directory with the new module name already exists: '{new_module_dir}'")
        sys.exit(1)

    # Copy Template Directory ---
    try:
        print(f"\n[1/6] Copying template directory...")
        shutil.copytree(template_dir, new_module_dir)
        print(f"  -> Successfully copied to '{new_module_dir}'")
    except OSError as e:
        print(f"\n[ERROR] Could not copy directory. Reason: {e}")
        sys.exit(1)

    # Rename Files and Replace Content ---
    # We walk through the newly created directory.
    print(f"\n[2/6] Renaming files and replacing content...")
    process_directory(new_module_dir, template_name, new_module_name)
    insert_c_declaration(component_file, new_module_name)

    # Generate UUID for the new module ---
    print("\n[3/6] Generating UUID for module...")
    insert_uuid_name(uuid_file, new_module_name)

    # Add CMake rule for new module ---
    print("\n[4/6] Module creation process finished successfully!")
    kconfig_option = f"CONFIG_COMP_{new_module_name.upper()}"
    print(f"  -> Adding CMake rule for '{kconfig_option}'")
    insert_cmake_rule(cmake_file, kconfig_option, new_module_name)

    # Add Kconfig rsource for new module
    print("\n[5/6] Module creation process finished successfully!")
    insert_kconfig_source(kconfig_file, f"{new_module_name}/Kconfig")

    print("\n[6/6] Module creation process finished successfully!")
    print("--- Done ---")

if __name__ == "__main__":
    main()




#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause

# Too much noise for now, these can be re-enabled after they've been
# fixed (if that does not break `git blame` too much)

# W0311, W0312, W0603
# pylint:disable=bad-indentation
# pylint:disable=mixed-indentation
# pylint:disable=global-statement

# C0103, C0114, C0116
# pylint:disable=invalid-name
# pylint:disable=missing-module-docstring
# pylint:disable=missing-function-docstring

# Non-indentation whitespace has been removed from newer pylint. It does
# not hurt to keep them for older versions. The recommendation is to use
# a formatter like `black` instead, unfortunately this would totally
# destroy git blame, git revert, etc.

# C0326, C0330
# pylint:disable=bad-whitespace
# pylint:disable=bad-continuation

import argparse
import shlex
import subprocess
import pathlib
import errno
import platform as py_platform
import sys
import shutil
import os
import warnings
import fnmatch
import hashlib
import json
import gzip
import dataclasses
import concurrent.futures as concurrent
import tarfile
import re

from west import configuration as west_config

# anytree module is defined in Zephyr build requirements
from anytree import AnyNode, RenderTree, render
from packaging import version

# https://chrisyeh96.github.io/2017/08/08/definitive-guide-python-imports.html#case-3-importing-from-parent-directory
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from tools.sof_ri_info import sof_ri_info

MIN_PYTHON_VERSION = 3, 8
assert sys.version_info >= MIN_PYTHON_VERSION, \
	f"Python {MIN_PYTHON_VERSION} or above is required."

# Version of this script matching Major.Minor.Patch style.
VERSION = version.Version("2.0.0")

# Constant value resolves SOF_TOP directory as: "this script directory/.."
SOF_TOP = pathlib.Path(__file__).parents[1].resolve()
west_top = pathlib.Path(SOF_TOP, "..").resolve()

sof_fw_version = None

signing_key = None

if py_platform.system() == "Windows":
	xtensa_tools_version_postfix = "-win32"
elif py_platform.system() == "Linux":
	xtensa_tools_version_postfix = "-linux"
else:
	xtensa_tools_version_postfix = "-unsupportedOS"
	warnings.warn(f"Your operating system: {py_platform.system()} is not supported")


@dataclasses.dataclass
# pylint:disable=too-many-instance-attributes
class PlatformConfig:
	"Product parameters"
	vendor: str
	PLAT_CONFIG: str
	XTENSA_TOOLS_VERSION: str
	XTENSA_CORE: str
	DEFAULT_TOOLCHAIN_VARIANT: str = "xt-clang"
	RIMAGE_KEY: pathlib.Path = pathlib.Path(SOF_TOP, "keys", "otc_private_key_3k.pem")
	aliases: list = dataclasses.field(default_factory=list)
	ipc4: bool = False

# These cannot be built by everyone out of the box yet.
# For instance: there's no open-source toolchain available for them yet.
extra_platform_configs = {
	"nvl" : PlatformConfig(
		"intel", "intel_adsp/ace40/nvl",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace4px_HiFi5MMU_PIF_nlib",
		ipc4 = True
	),
	"nvl-s" : PlatformConfig(
		"intel", "intel_adsp/ace40/nvls",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace4px_HiFi5MMU_PIF_nlib",
		ipc4 = True
	),
	"wcl-sim" : PlatformConfig(
		"intel", "intel_adsp/ace30/wcl/sim",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace30_LX7HiFi4_PIF",
		ipc4 = True
	),
	"ptl-sim" : PlatformConfig(
		"intel", "intel_adsp/ace30/ptl/sim",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace30_LX7HiFi4_PIF",
		ipc4 = True
	),
	# AMD platforms
	"acp_6_0" : PlatformConfig(
		"amd", "acp_6_0_adsp/acp_6_0",
		f"RI-2019.1{xtensa_tools_version_postfix}",
		"rmb_LX7_HiFi5_PROD",
		RIMAGE_KEY = "key param ignored by acp_6_0"
	),
	# MediaTek platforms
	# (move to platform_configs_all on next Zephyr SDK release after 0.17.0)
	"mt8195" : PlatformConfig(
		"mtk", "mt8195/mt8195/adsp",
		f"RJ-2024.3{xtensa_tools_version_postfix}",
		"hifi4_8195_PROD",
	),
	"mt8186" : PlatformConfig(
		"mtk", "mt8186/mt8186/adsp",
		f"RJ-2024.3{xtensa_tools_version_postfix}",
		"hifi5_7stg_I64D128",
	),
	"mt8188" : PlatformConfig(
		"mtk", "mt8188/mt8188/adsp",
		f"RJ-2024.3{xtensa_tools_version_postfix}",
		"hifi5_7stg_I64D128",
	),
	"mt8196" : PlatformConfig(
		"mtk", "mt8196/mt8196/adsp",
		f"RJ-2024.3{xtensa_tools_version_postfix}",
		"HiFi5_MPU_lock_2023_11",
	),
}

# These can all be built out of the box. --all builds all these.
# Some of these values are duplicated in sof/scripts/set_xtensa_param.sh: keep them in sync.
platform_configs_all = {
	#  Intel platforms
	"tgl" : PlatformConfig(
		"intel", "intel_adsp/cavs25",
		f"RG-2017.8{xtensa_tools_version_postfix}",
		"cavs2x_LX6HiFi3_2017_8",
		"xcc",
		aliases = ['adl', 'adl-n', 'rpl'],
		ipc4 = True
	),
	"tgl-h" : PlatformConfig(
		"intel", "intel_adsp/cavs25/tgph",
		f"RG-2017.8{xtensa_tools_version_postfix}",
		"cavs2x_LX6HiFi3_2017_8",
		"xcc",
		aliases = ['adl-s', 'rpl-s'],
		ipc4 = True
	),
	"mtl" : PlatformConfig(
		"intel", "intel_adsp/ace15_mtpm",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace10_LX7HiFi4_2022_10",
		aliases = ['arl', 'arl-s'],
		ipc4 = True
	),
	"lnl" : PlatformConfig(
		"intel", "intel_adsp/ace20_lnl",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace10_LX7HiFi4_2022_10",
		ipc4 = True
	),
	"ptl" : PlatformConfig(
		"intel", "intel_adsp/ace30/ptl",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace30_LX7HiFi4_PIF",
		ipc4 = True
	),
	"wcl" : PlatformConfig(
		"intel", "intel_adsp/ace30/wcl",
		f"RI-2022.10{xtensa_tools_version_postfix}",
		"ace30_LX7HiFi4_PIF",
		ipc4 = True
	),

	#  NXP platforms
	"imx8" : PlatformConfig(
		"imx", "imx8qm_mek/mimx8qm6/adsp",
		f"RI-2023.11{xtensa_tools_version_postfix}",
		"hifi4_nxp_v5_3_1_prod",
		RIMAGE_KEY = "key param ignored by imx8",
	),
	"imx8x" : PlatformConfig(
		"imx", "imx8qxp_mek/mimx8qx6/adsp",
		f"RI-2023.11{xtensa_tools_version_postfix}",
		"hifi4_nxp_v5_3_1_prod",
		RIMAGE_KEY = "key param ignored by imx8x"
	),
	"imx8m" : PlatformConfig(
		"imx", "imx8mp_evk/mimx8ml8/adsp",
		f"RI-2023.11{xtensa_tools_version_postfix}",
		"hifi4_mscale_v2_0_2_prod",
		RIMAGE_KEY = "key param ignored by imx8m"
	),
	"imx8ulp" : PlatformConfig(
		"imx", "imx8ulp_evk/mimx8ud7/adsp",
		f"RI-2023.11{xtensa_tools_version_postfix}",
		"hifi4_nxp2_s7_v2_1a_prod",
		RIMAGE_KEY = "key param ignored by imx8ulp"
	),
	"imx95" : PlatformConfig(
		"imx", "imx95_evk/mimx9596/m7/ddr",
		"", "", "", ""
	),
}

platform_configs = platform_configs_all.copy()
platform_configs.update(extra_platform_configs)

class validate_platforms_arguments(argparse.Action):
	"""Validates positional platform arguments whether provided platform name is supported."""
	def __call__(self, parser, namespace, values, option_string=None):
		if values:
			for value in values:
				if value not in platform_configs:
					raise argparse.ArgumentError(self, f"Unsupported platform: {value}")
		setattr(namespace, "platforms", values)

args = None
def parse_args():
	global args
	global west_top
	parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
			epilog=(

f"""This script supports XtensaTools but only when installed in a specific directory
structure where a single --xtensa-core is registered per --xtensa-system registry, example
below. Different layouts may or may not work.

myXtensa/
└── install/
       ├── builds/
       │   ├── RD-2012.5{xtensa_tools_version_postfix}
       │   │   └── Intel_HiFiEP/config/        # XTENSA_SYSTEM registry
       │   │        ├── default-params
       │   │        └── Intel_HiFiEP-params    # single, default XTENSA_CORE per registry
       │   └── RG-2017.8{xtensa_tools_version_postfix}
       │       ├── LX4_langwell_audio_17_8/config/
       │       │    ├── default-params
       │       │    └── LX4_langwell_audio_17_8-params
       │       └── X4H3I16w2D48w3a_2017_8/config/
       │           ├── default-params
       │           └── X4H3I16w2D48w3a_2017_8-params
       └── tools/
               ├── RD-2012.5{xtensa_tools_version_postfix}
               │   └── XtensaTools
               └── RG-2017.8{xtensa_tools_version_postfix}
                   └── XtensaTools/

$XTENSA_TOOLS_ROOT=/path/to/myXtensa ...

Supported platforms: {list(platform_configs)}"""

			), add_help=False)

	parser.add_argument('-h', '--help', action='store_true', help='show help')
	parser.add_argument("-a", "--all", required=False, action="store_true",
						help="Build all currently supported platforms")
	parser.add_argument("platforms", nargs="*", action=validate_platforms_arguments,
			    help="List of platforms to build. Zero platform + any flag (-p) still creates 'tools/'")
	parser.add_argument("-d", "--debug", required=False, action="store_true",
						help="Shortcut for: -o sof/app/debug_overlay.conf")
    # NO SOF release will ever user the option --fw-naming.
    # This option is only for disguising SOF IPC4 as CAVS IPC4 and only in cases where
    # the kernel 'ipc_type' expects CAVS IPC4. In this way, developers and CI can test
    # IPC4 on older platforms.
	parser.add_argument("--fw-naming", required=False, choices=["AVS", "SOF"],
						default="SOF",
help="""Determine firmware naming conversion and folder structure.
See also the newer and better '--deployable-build' for IPC4.
With "SOF" (default):

build-sof-staging (for /lib/firmware/intel/sof/)
        ├───community
        │   └── sof-tgl.ri
        ├── dbgkey
        │   └── sof-tgl.ri
        ├── sof-tgl.ri
        └── sof-tgl.ldc

With "AVS"; filename 'dsp_basefw.bin'.
'AVS' automatically enables --use-platform-subdir which uses one subdirectory for each platform:

build-sof-staging (for old, developer /lib/firmware/intel/sof-ipc4/)
    └── tgl
        ├── community
        │   └── dsp_basefw.bin
        ├── dbgkey
        │   └── dsp_basefw.bin
        ├── dsp_basefw.bin
        └── sof-tgl.ldc

""")
	parser.add_argument("-j", "--jobs", required=False, type=int,
						help="Number of concurrent jobs. Passed to west build and"
						" to cmake (for rimage)")
	parser.add_argument("-k", "--key", type=pathlib.Path, required=False,
						help="Path to a non-default rimage signing key.")

	# https://docs.zephyrproject.org/latest/build/kconfig/setting.html#initial-conf
	# https://docs.zephyrproject.org/latest/develop/application/index.html#important-build-system-variables
	# TODO, support "snippets"?
	# https://docs.zephyrproject.org/latest/build/snippets/writing.html
	parser.add_argument("-o", "--overlay", type=pathlib.Path, required=False, action='append',
			default=[], help=
"""All '-o arg1 -o arg2 ...' arguments are combined into a single -DEXTRA_CONF_FILE='arg1;arg2;...'
list. This tweaks the build CONFIGuration like Device Tree "overlays" and Zephyr snippets do.
Files latter in the list seem to have precedence. Direct -C=-DCONFIG_xxx=.. options seem to
have precedence over -DEXTRA_CONF_FILE=... Rely on precedence as little as possible.""")

	parser.add_argument("-p", "--pristine", required=False, action="store_true",
						help="Perform pristine build removing build directory.")
	parser.add_argument("-u", "--update", required=False, action="store_true",
		help="""Runs west update command - clones SOF dependencies. Downloads next to this sof clone a new Zephyr
project with its required dependencies. Creates a modules/audio/sof symbolic link pointing
back at this sof clone.  All projects are checkout out to
revision defined in manifests of SOF and Zephyr.""")
	parser.add_argument('-v', '--verbose', default=0, action='count',
			    help="""Verbosity level. Repetition of the flag increases verbosity.
The same number of '-v' is passed to "west".
""",
	)
	# Cannot use a standard -- delimiter because argparse deletes it.
	parser.add_argument("-C", "--cmake-args", action='append', default=[],
			    help="""Cmake arguments passed as is to cmake configure step.
Can be passed multiple times; whitespace is preserved Example:

     -C=--warn-uninitialized  -C '-DEXTRA_FLAGS=-Werror -g0'

Note '-C --warn-uninitialized' is not supported by argparse, an equal
sign must be used (https://bugs.python.org/issue9334)""",
	)

	parser.add_argument("--key-type-subdir", default="community",
			    choices=["community", "none", "dbgkey"],
			    help="""Output subdirectory for rimage signing key type.
Default key type subdirectory is \"community\".""")


	parser.add_argument("--use-platform-subdir", default = False,
			    action="store_true",
			    help="""Use an output subdirectory for each platform.
Otherwise, all firmware files are installed in the same staging directory by default.""")

	parser.add_argument("--no-interactive", default=False, action="store_true",
			    help="""Run script in non-interactive mode when user input can not be provided.
This should be used with programmatic script invocations (eg. Continuous Integration).
				""")

	deploy_args = parser.add_mutually_exclusive_group()
	# argparse.BooleanOptionalAction requires Python 3.9
	parser.set_defaults(deployable_build=True)
	deploy_args.add_argument("--no-deployable-build", dest='deployable_build', action='store_false')
	deploy_args.add_argument("--deployable-build", dest='deployable_build', action='store_true',
			    help="""Create a directory structure for the firmware files which can be deployed on target as it is.
This option will cause the --fw-naming and --use-platform-subdir options to be ignored!
The generic, default directory and file structure is IPC version dependent:
IPC3
    build-sof-staging/sof/VENDOR/sof/ (on target: /lib/firmware/VENDOR/sof/)
    ├── community
    │   └── sof-PLAT.ri
    ├── dbgkey
    │   └── sof-PLAT.ri
    └── sof-PLAT.ri
IPC4
    build-sof-staging/sof/VENDOR/sof-ipc4/ (on target: /lib/firmware/VENDOR/sof-ipc4/)
    └── PLAT
        ├── community
        │   └── sof-PLAT.ri
        ├── dbgkey
        │   └── sof-PLAT.ri
        └── sof-PLAT.ri\n
				""")

	parser.add_argument("--no-tarball", default=False, action='store_true',
			    help="""Do not create a tar file of the installed firmware files under build-sof-staging.""")
	parser.add_argument("--version", required=False, action="store_true",
			    help="Prints version of this script.")
	parser.add_argument("-m", "--menuconfig", required=False, action="store_true",
				help="Build menuconfig for target")
	parser.add_argument("-z", "--zephyrsdk", required=False, action="store_true",
				help="Force Build using Zephyr SDK for target")

	args = parser.parse_args()

	# This does NOT include "extra", experimental platforms.
	if args.all:
		args.platforms = list(platform_configs_all)

	# print help message if no arguments provided
	if len(sys.argv) == 1 or args.help:
		for platform in platform_configs:
			if platform_configs[platform].aliases:
				parser.epilog += "\nPlatform aliases for '" + platform + "':\t"
				parser.epilog += f"{list(platform_configs[platform].aliases)}"

		parser.print_help()
		sys.exit(0)

	if args.deployable_build:
		if args.fw_naming == 'AVS' or args.use_platform_subdir:
			sys.exit("Options '--fw-naming=AVS' and '--use-platform-subdir'"
				 " are incompatible with deployable builds, try --no-deployable-build?")

	if args.fw_naming == 'AVS':
		if not args.use_platform_subdir:
			args.use_platform_subdir=True
			warnings.warn("The option '--fw-naming AVS' has to be used with '--use-platform-subdir'. Enable '--use-platform-subdir' automatically.")

def execute_command(*run_args, **run_kwargs):
	"""[summary] Provides wrapper for subprocess.run that prints
	command executed when 'more verbose' verbosity level is set."""
	command_args = run_args[0]

	# If you really need the shell in some non-portable section then
	# invoke subprocess.run() directly.
	if run_kwargs.get('shell') or not isinstance(command_args, list):
		raise RuntimeError("Do not rely on non-portable shell parsing")

	if args.verbose >= 0:
		cwd = run_kwargs.get('cwd')
		print_cwd = f"In dir: {cwd}" if cwd else f"in current dir: {os.getcwd()}"
		print_args = shlex.join(command_args)
		output = f"{print_cwd}; running command:\n    {print_args}"
		env_arg = run_kwargs.get('env')
		env_change = set(env_arg.items()) - set(os.environ.items()) if env_arg else None
		if env_change and (run_kwargs.get('sof_log_env') or args.verbose >= 1):
			output += "\n... with extra/modified environment:\n"
			for k_v in env_change:
				output += f"{k_v[0]}={k_v[1]}\n"
		print(output, flush=True)

	run_kwargs = {k: run_kwargs[k] for k in run_kwargs if not k.startswith("sof_")}

	if not 'check' in run_kwargs:
		run_kwargs['check'] = True
	#pylint:disable=subprocess-run-check

	return subprocess.run(*run_args, **run_kwargs)

def symlink_or_copy(targetdir, targetfile, linkdir, linkfile):
	"""Create a relative symbolic link or copy. Don't bother Windows users
	with symbolic links because they require special privileges.
	Windows don't care about /lib/firmware/sof/ anyway.
	Make a copy instead to preserve cross-platform consistency.
	"""
	target = pathlib.Path(targetdir) / targetfile
	link = pathlib.Path(linkdir) / linkfile
	if py_platform.system() == "Windows":
		shutil.copy2(target, link)
	else:
		link.symlink_to(os.path.relpath(target, linkdir))

def show_installed_files():
	"""[summary] Scans output directory building binary tree from files and folders
	then presents them in similar way to linux tree command."""
	graph_root = AnyNode(name=STAGING_DIR.name, long_name=".", parent=None)
	relative_entries = [
		entry.relative_to(STAGING_DIR) for entry in sorted(STAGING_DIR.glob("**/*"))
	]
	nodes = [ graph_root ]
	for entry in relative_entries:
		# Node's documentation does allow random attributes
		# pylint: disable=no-member
		# sorted() makes sure our parent is already there.
		# This is slightly awkward, a recursive function would be more readable
		matches = [node for node in nodes if node.long_name == str(entry.parent)]
		assert len(matches) == 1, f'"{entry}" does not have exactly one parent'
		nodes.append(AnyNode(name=entry.name, long_name=str(entry), parent=matches[0]))

	for pre, _, node in RenderTree(graph_root, render.AsciiStyle):
		fpath = STAGING_DIR / node.long_name

		# pathLib.readlink() requires Python 3.9
		symlink_trailer = f" -> {os.readlink(fpath)}" if fpath.is_symlink() else ""

		stem = node.name[:-3] if node.name.endswith(".gz") else node.name

		shasum_trailer = ""
		if checksum_wanted(stem) and fpath.is_file() and not fpath.is_symlink():
			shasum_trailer =  "\tsha256=" + checksum(fpath)

		print(f"{pre}{node.name} {symlink_trailer} {shasum_trailer}")


# TODO: among other things in this file it should be less SOF-specific;
# try to move as much as possible to generic Zephyr code. See
# discussions in https://github.com/zephyrproject-rtos/zephyr/pull/51954
def checksum_wanted(stem):
	for pattern in CHECKSUM_WANTED:
		if fnmatch.fnmatch(stem, pattern):
			return True
	return False


def checksum(fpath):
	if fpath.suffix == ".gz":
		inputf = gzip.GzipFile(fpath, "rb")
	else:
		inputf = open(fpath, "rb")
	chksum = hashlib.sha256(inputf.read()).hexdigest()
	inputf.close()
	return chksum


def check_west_installation():
	west_path = shutil.which("west")
	if not west_path:
		raise FileNotFoundError("Install west and a west toolchain,"
			"https://docs.zephyrproject.org/latest/getting_started/index.html")
	print(f"Found west: {west_path}")

def west_reinitialize(west_root_dir: pathlib.Path, west_manifest_path: pathlib.Path):
	"""[summary] Performs west reinitialization to SOF manifest file asking user for permission.
	Prints error message if script is running in non-interactive mode.

	:param west_root_dir: directory where is initialized.
	:type west_root_dir: pathlib.Path
	:param west_manifest_path: manifest file to which west is initialized.
	:type west_manifest_path: pathlib.Path
	:raises RuntimeError: Raised when west is initialized to wrong manifest file
	(not SOFs manifest) and script is running in non-interactive mode.
	"""
	global west_top
	message = "West is initialized to manifest other than SOFs!\n"
	message +=  f"Initialized to manifest: {west_manifest_path}." + "\n"
	dot_west_directory  = pathlib.Path(west_root_dir.resolve(), ".west")
	if args.no_interactive:
		message += f"Try deleting {dot_west_directory } directory and rerun this script."
		raise RuntimeError(message)
	question = message + "Reinitialize west to SOF manifest? [Y/n] "
	print(f"{question}")
	while True:
		reinitialize_answer = input().lower()
		if reinitialize_answer in ["y", "n"]:
			break
		sys.stdout.write('Please respond with \'Y\' or \'n\'.\n')

	if reinitialize_answer != 'y':
		sys.exit("Can not proceed. Reinitialize your west manifest to SOF and rerun this script.")
	shutil.rmtree(dot_west_directory)
	execute_command(["west", "init", "-l", f"{SOF_TOP}"], cwd=west_top)


# TODO: use west APIs directly instead of all these indirect subprocess.run("west", ...) processes
def west_init_if_needed():
	"""[summary] Validates whether west workspace had been initialized and points to SOF manifest.
	Peforms west initialization if needed.
	"""
	global west_top, SOF_TOP
	west_manifest_path = pathlib.Path(SOF_TOP, "west.yml")
	result_rootdir = execute_command(["west", "topdir"], capture_output=True, cwd=west_top,
		timeout=10, check=False)
	if result_rootdir.returncode != 0:
		execute_command(["west", "init", "-l", f"{SOF_TOP}"], cwd=west_top)
		return
	west_root_dir = pathlib.Path(result_rootdir.stdout.decode().rstrip()).resolve()
	result_manifest_dir = execute_command(["west", "config", "manifest.path"], capture_output=True,
		cwd=west_top, timeout=10, check=True)
	west_manifest_dir = pathlib.Path(west_root_dir, result_manifest_dir.stdout.decode().rstrip()).resolve()
	manifest_file_result = execute_command(["west", "config", "manifest.file"], capture_output=True,
		cwd=west_top, timeout=10, check=True)
	returned_manifest_path = pathlib.Path(west_manifest_dir, manifest_file_result.stdout.decode().rstrip())
	if not returned_manifest_path.samefile(west_manifest_path):
		west_reinitialize(west_root_dir, returned_manifest_path)
	else:
		print(f"West workspace: {west_root_dir}")
		print(f"West manifest path: {west_manifest_path}")

def create_zephyr_directory():
	global west_top
	# Do not fail when there's only an empty directory left over
	# (because of some early interruption of this script or proxy
	# misconfiguration, etc.)
	try:
		# rmdir() is safe: it deletes empty directories ONLY.
		west_top.rmdir()
	except OSError as oserr:
		if oserr.errno not in [errno.ENOTEMPTY, errno.ENOENT]:
			raise oserr
		# else when not empty then let the next line fail with a
		# _better_ error message:
		#         "zephyrproject already exists"

	west_top.mkdir(parents=False, exist_ok=False)
	west_top = west_top.resolve(strict=True)

def create_zephyr_sof_symlink():
	global west_top, SOF_TOP
	if not west_top.exists():
		raise FileNotFoundError("No west top: {}".format(west_top))
	audio_modules_dir = pathlib.Path(west_top, "modules", "audio")
	audio_modules_dir.mkdir(parents=True, exist_ok=True)
	sof_symlink = pathlib.Path(audio_modules_dir, "sof")
	# Symlinks creation requires administrative privileges in Windows or special user rights
	try:
		if not sof_symlink.exists():
			sof_symlink.symlink_to(SOF_TOP, target_is_directory=True)
	except:
		print(f"Failed to create symbolic link: {sof_symlink} to {SOF_TOP}."
			"\nIf you run script on Windows run it with administrative privileges or\n"
			"grant user symlink creation rights -"
			"see: https://docs.microsoft.com/en-us/windows/security/threat-protection/"
			"security-policy-settings/create-symbolic-links")
		raise

def west_update():
	"""[summary] Clones all west manifest projects to specified revisions"""
	global west_top
	execute_command(["west", "update"], check=True, timeout=3000, cwd=west_top)


def get_sof_version():
	"""[summary] Get version string major.minor.micro of SOF
	firmware file. When building multiple platforms from the same SOF
	commit, all platforms share the same version. So for the 1st platform,
	extract the version information from sof/versions.json and later platforms will
	reuse it.
	"""
	global sof_fw_version
	if sof_fw_version:
		return sof_fw_version

	versions = {}
	with open(SOF_TOP / "versions.json") as versions_file:
		versions = json.load(versions_file)
		# Keep this default value the same as the default SOF_MICRO in version.cmake
		sof_micro = versions['SOF'].get('MICRO', "0")
		sof_fw_version = (
			f"{versions['SOF']['MAJOR']}.{versions['SOF']['MINOR']}.{sof_micro}"
		)
	return sof_fw_version

def rmtree_if_exists(directory):
	"This is different from ignore_errors=False because it deletes everything or nothing"
	if os.path.exists(directory):
		shutil.rmtree(directory)

def clean_staging(platform):
	print(f"Cleaning {platform} from {STAGING_DIR}")

	rmtree_if_exists(STAGING_DIR / "sof-info" / platform)

	sof_output_dir = STAGING_DIR / "sof"

	# --use-platform-subdir
	rmtree_if_exists(sof_output_dir / platform)

	# Remaining .ri and .ldc files
	for p in [ platform ] + platform_configs[platform].aliases:
		for f in sof_output_dir.glob(f"**/sof-{p}.*"):
			os.remove(f)

	# remove IPC4 deployable build directories
	if platform_configs[platform].ipc4:
		# e.g.: "build-sof-staging/sof/intel/sof-ipc4/mtl/"
		sof_vendor_dir = sof_output_dir / platform_configs[platform].vendor
		rmtree_if_exists(sof_vendor_dir / "sof-ipc4" / platform)
		rmtree_if_exists(sof_vendor_dir / "sof-ipc4-lib" / platform)
		for p_alias in platform_configs[platform].aliases:
			rmtree_if_exists(sof_vendor_dir / "sof-ipc4" / p_alias)
			rmtree_if_exists(sof_vendor_dir / "sof-ipc4-lib" / p_alias)

RIMAGE_BUILD_DIR  = west_top / "build-rimage"
RIMAGE_SOURCE_DIR = west_top / "sof" / "tools" / "rimage"


def rimage_west_configuration(platform_dict, dest_dir):
	"""Configure rimage in a new file `dest_dir/westconfig.ini`, starting
	from the workspace .west/config.
	Returns a tuple (west ConfigFile, pathlib.Path to that new file).
	"""

	saved_local_var = os.environ.get('WEST_CONFIG_LOCAL')
	workspace_west_config_path = os.environ.get('WEST_CONFIG_LOCAL',
						   str(west_top / ".west" / "config"))
	platform_west_config_path = dest_dir / "westconfig.ini"
	dest_dir.mkdir(parents=True, exist_ok=True)
	shutil.copyfile(workspace_west_config_path, platform_west_config_path)

	# Create `platform_wconfig` object pointing at our copy
	os.environ['WEST_CONFIG_LOCAL'] = str(platform_west_config_path)
	platform_wconfig = west_config.Configuration()
	if saved_local_var is None:
		del os.environ['WEST_CONFIG_LOCAL']
	else:
		os.environ['WEST_CONFIG_LOCAL'] = saved_local_var

	# By default, run rimage directly from the rimage build directory
	if platform_wconfig.get("rimage.path") is None:
		rimage_executable = shutil.which("rimage", path=RIMAGE_BUILD_DIR)
		assert pathlib.Path(str(rimage_executable)).exists()
		platform_wconfig.set("rimage.path", shlex.quote(rimage_executable),
				     west_config.ConfigFile.LOCAL)

	_ws_args = platform_wconfig.get("rimage.extra-args")
	workspace_extra_args = [] if _ws_args is None else shlex.split(_ws_args)

	# Flatten default rimage options while giving precedence to the workspace =
	# the user input. We could just append and leave duplicates but that would be
	# at best confusing and at worst relying on undocumented rimage precedence.
	extra_args = []
	for default_opt in rimage_options(platform_dict):
		if not default_opt[0] in workspace_extra_args:
			extra_args += default_opt

	extra_args += workspace_extra_args

	platform_wconfig.set("rimage.extra-args", shlex.join(extra_args))

	return platform_wconfig, platform_west_config_path


def build_rimage():

	old_rimage_loc = SOF_TOP / "rimage"
	# Don't warn on empty directories
	if ( old_rimage_loc  / "CMakeLists.txt" ).exists():
		warnings.warn(f"""{old_rimage_loc} is now ignored,
		new location is {RIMAGE_SOURCE_DIR}"""
		)
	rimage_dir_name = RIMAGE_BUILD_DIR.name
	# CMake build rimage module
	if not (RIMAGE_BUILD_DIR / "CMakeCache.txt").is_file():
		execute_command(["cmake", "-B", rimage_dir_name, "-G", "Ninja",
				 "-S", str(RIMAGE_SOURCE_DIR)],
				cwd=west_top)
	rimage_build_cmd = ["cmake", "--build", rimage_dir_name]
	if args.jobs is not None:
		rimage_build_cmd.append(f"-j{args.jobs}")
	if args.verbose > 1:
			rimage_build_cmd.append("-v")
	execute_command(rimage_build_cmd, cwd=west_top)


def rimage_options(platform_dict):
	"""Return a list of default rimage options as a list of tuples,
	example: [ (-f, 2.5.0), (-b, 1), (-k, key.pem),... ]

	"""
	global signing_key
	opts = []

	if args.verbose > 0:
		opts.append(("-v",) * args.verbose)

	if args.key:
		key_path = pathlib.Path(args.key)
		assert key_path.exists(), f"{key_path} not found"
		signing_key = key_path.resolve()
	elif "RIMAGE_KEY" in platform_dict:
		signing_key = platform_dict["RIMAGE_KEY"]

	if signing_key is not None:
		opts.append(("-k", str(signing_key)))

	sof_fw_vers = get_sof_version()

	opts.append(("-f", sof_fw_vers))

	# Default value is 0 in rimage but for Zephyr the "build counter" has always
	# been hardcoded to 1 in CMake and there is even a (broken) test that fails
	# when it's not hardcoded to 1.
	# FIXME: drop this line once the following test is fixed
	# tests/avs/fw_00_basic/test_01_load_fw_extended.py::TestLoadFwExtended::()::
	#                         test_00_01_load_fw_and_check_version
	opts.append(("-b", "1"))

	return opts


STAGING_DIR = None
def build_platforms():
	global west_top, SOF_TOP
	print(f"SOF_TOP={SOF_TOP}")
	print(f"west_top={west_top}")

	global STAGING_DIR
	STAGING_DIR = pathlib.Path(west_top, "build-sof-staging")
	# Don't leave the install of an old build behind
	if args.pristine:
		rmtree_if_exists(STAGING_DIR)
	else:
		# This is important in (at least) two use cases:
		# - when switching `--use-platform-subdir` on/off or changing key subdir,
		# - when the build starts failing after a code change.
		# Do not delete platforms that were not requested so this script can be
		# invoked once per platform.
		for platform in args.platforms:
			clean_staging(platform)
		rmtree_if_exists(STAGING_DIR / "tools")


	# smex does not use 'install -D'
	sof_output_dir = pathlib.Path(STAGING_DIR, "sof")
	sof_output_dir.mkdir(parents=True, exist_ok=True)

	platform_build_dir_name = None
	for platform in args.platforms:
		platf_build_environ = os.environ.copy()

		if args.deployable_build:
			vendor_output_dir = pathlib.Path(sof_output_dir, platform_configs[platform].vendor)
			if platform_configs[platform].ipc4:
				sof_platform_output_dir = pathlib.Path(vendor_output_dir, "sof-ipc4")
			else:
				sof_platform_output_dir = pathlib.Path(vendor_output_dir, "sof")

			sof_platform_output_dir.mkdir(parents=True, exist_ok=True)
		else:
			if args.use_platform_subdir:
				sof_platform_output_dir = pathlib.Path(sof_output_dir, platform)
				sof_platform_output_dir.mkdir(parents=True, exist_ok=True)
			else:
				sof_platform_output_dir = sof_output_dir

		# For now convert the new dataclass to what it used to be
		_dict = dataclasses.asdict(platform_configs[platform])
		platform_dict = { k:v for (k,v) in _dict.items() if _dict[k] is not None }

		if args.zephyrsdk:
			print("Using Zephyr SDK for building")
			xtensa_tools_root_dir = None
		else:
			print("Using Xtensa tools for building")
			xtensa_tools_root_dir = os.getenv("XTENSA_TOOLS_ROOT")
		# when XTENSA_TOOLS_ROOT environmental variable is set,
		# use user installed Xtensa tools not Zephyr SDK
		if "XTENSA_TOOLS_VERSION" in platform_dict and xtensa_tools_root_dir:
			xtensa_tools_root_dir = pathlib.Path(xtensa_tools_root_dir)
			if not xtensa_tools_root_dir.is_dir():
				raise RuntimeError(f"Platform {platform} uses Xtensa toolchain."
					f"\nVariable XTENSA_TOOLS_ROOT={xtensa_tools_root_dir} points "
					"to a path that does not exist or is not a directory")

			# Set environment variables expected by
			# zephyr/cmake/toolchain/x*/*.cmake.  Note CMake cannot set (evil)
			# build-time environment variables at configure time:
			# https://gitlab.kitware.com/cmake/community/-/wikis/FAQ#how-can-i-get-or-set-environment-variables

			# Top-level "switch". When undefined, the Zephyr build system
			# automatically searches for the Zephyr SDK and ignores all the
			# following variables.
			platf_build_environ["ZEPHYR_TOOLCHAIN_VARIANT"] = platf_build_environ.get("ZEPHYR_TOOLCHAIN_VARIANT",
				platform_dict["DEFAULT_TOOLCHAIN_VARIANT"])
			platf_build_environ["XTENSA_TOOLCHAIN_PATH"] = str(
				xtensa_tools_root_dir / "install" / "tools"
			)
			# Toolchain sub-directory
			TOOLCHAIN_VER = platform_dict["XTENSA_TOOLS_VERSION"]
			platf_build_environ["TOOLCHAIN_VER"] = TOOLCHAIN_VER

			# This XTENSA_SYSTEM variable was copied as is from XTOS
			# scripts/xtensa-build-all.sh but it is not required by (recent?)
			# toolchains when installed exactly as described in the --help (= a
			# single --xtensa-core installed per --xtensa-system registry and maybe
			# other constraints). Let's keep exporting it in case some build
			# systems have their toolchains installed differently which might
			# require it.  Note: "not needed" is different from "ignored"! The
			# compiler fails when the value is incorrect.
			builds_toolchainver = xtensa_tools_root_dir / "install" / "builds" / TOOLCHAIN_VER
			XTENSA_SYSTEM = str(builds_toolchainver / platform_dict["XTENSA_CORE"] / "config")
			platf_build_environ["XTENSA_SYSTEM"] = XTENSA_SYSTEM

		platform_build_dir_name = f"build-{platform}"

		PLAT_CONFIG = platform_dict["PLAT_CONFIG"]
		build_cmd = ["west"]
		build_cmd += ["-v"] * args.verbose
		if args.menuconfig:
			build_cmd += ["build", "-t", "menuconfig", "--build-dir", platform_build_dir_name]
		else:
			build_cmd += ["build", "--build-dir", platform_build_dir_name]
		source_dir = pathlib.Path(SOF_TOP, "app")
		build_cmd += ["--board", PLAT_CONFIG, str(source_dir)]
		if args.pristine:
			build_cmd += ["-p", "always"]

		if args.jobs is not None:
			build_cmd += [f"--build-opt=-j{args.jobs}"]

		build_cmd.append('--')
		if args.cmake_args:
			build_cmd += args.cmake_args

		extra_conf_files = [str(item.resolve(True)) for item in args.overlay]
		# The '-d' option is a shortcut for '-o path_to_debug_overlay', we are good
		# if both are provided, because it's no harm to merge the same overlay twice.
		if args.debug:
			extra_conf_files.append(str(pathlib.Path(SOF_TOP, "app", "debug_overlay.conf")))

		# The xt-cland Cadence toolchain currently cannot link shared
		# libraries for Xtensa. Therefore when it's used we switch to
		# building relocatable ELF objects.
		if platf_build_environ.get("ZEPHYR_TOOLCHAIN_VARIANT") == 'xt-clang':
			extra_conf_files.append(str(pathlib.Path(SOF_TOP, "app", "llext_relocatable.conf")))

		if extra_conf_files:
			extra_conf_files = ";".join(extra_conf_files)
			build_cmd.append(f"-DEXTRA_CONF_FILE={extra_conf_files}")

		abs_build_dir = pathlib.Path(west_top, platform_build_dir_name)

		# Longer story in https://github.com/zephyrproject-rtos/zephyr/pull/56671
		if not args.pristine and (
			pathlib.Path(abs_build_dir, "build.ninja").is_file()
			or pathlib.Path(abs_build_dir, "Makefile").is_file()
		):
			if args.cmake_args or extra_conf_files:
				warnings.warn("""CMake args slow down incremental builds.
	Passing CMake parameters and -o conf files on the command line slows down incremental builds
	see https://docs.zephyrproject.org/latest/guides/west/build-flash-debug.html#one-time-cmake-arguments
	Try "west config build.cmake-args -- ..." instead.""")

		platform_wcfg, wcfg_path = rimage_west_configuration(
			platform_dict,
			STAGING_DIR / "sof-info" / platform
		)
		platf_build_environ['WEST_CONFIG_LOCAL'] = str(wcfg_path)

		# Make sure the build logs don't leave anything hidden
		execute_command(['west', 'config', '-l'], cwd=west_top,
				env=platf_build_environ, sof_log_env=True)
		print()

		# Build
		try:
			execute_command(build_cmd, cwd=west_top, env=platf_build_environ, sof_log_env=True)
		except subprocess.CalledProcessError as cpe:
			zephyr_path = pathlib.Path(west_top, "zephyr")
			if not os.path.exists(zephyr_path):
				sys.exit("Zephyr project not found. Please run this script with -u flag or `west update zephyr` manually.")
			else: # unknown failure
				raise cpe

		# no output to display if building menuconfig
		if args.menuconfig:
			return

		# reproducible-zephyr.ri is less useful now that show_installed_files() shows
		# checksums too. However: - it's still useful when only the .ri file is
		# available (no build logs for the other image), - it makes sure sof_ri_info.py
		# itself does not bitrot, - it can catch rimage issues, see for instance rimage
		# fix 4fb9fe00575b
		if platform not in RI_INFO_UNSUPPORTED:
			reproducible_checksum(platform, west_top / platform_build_dir_name / "zephyr" / "zephyr.ri")

		install_platform(platform, sof_platform_output_dir, platf_build_environ, platform_wcfg)

	src_dest_list = []
	tools_output_dir = pathlib.Path(STAGING_DIR, "tools")

	src_dest_list += [(pathlib.Path(SOF_TOP) /
		"tools" / "mtrace"/ "mtrace-reader.py",
		tools_output_dir / "mtrace-reader.py")]

	# Append future files to `src_dest_list` here (but prefer
	# copying entire directories; more flexible)

	for _src, _dst in src_dest_list:
		os.makedirs(os.path.dirname(_dst), exist_ok=True)
		# looses file owner and group - file is commonly accessible
		shutil.copy2(str(_src), str(_dst))

	# cavstool and friends
	shutil.copytree(pathlib.Path(west_top) /
			  "zephyr" / "soc" / "intel" / "intel_adsp" / "tools",
			tools_output_dir,
			symlinks=True, ignore_dangling_symlinks=True, dirs_exist_ok=True)


def install_lib(platform, sof_output_dir, abs_build_dir, platform_wconfig):
	"""[summary] Sign loadable llext modules, if any, copy them to the
	deployment tree and create UUID links for the kernel to find and load
	them."""

	global signing_key

	libs = dict()
	lib_uuids = dict()
	rimage_cmd = shlex.split(platform_wconfig.get('rimage.path'))[0]
	_ws_args = platform_wconfig.get("rimage.extra-args")

	sof_lib_dir = sof_output_dir / '..' / 'sof-ipc4-lib' / platform

	if args.key_type_subdir != "none":
		sof_lib_dir = sof_lib_dir / args.key_type_subdir

	sof_lib_dir.mkdir(parents=True, exist_ok=True)

	dirlist = [d for d in os.listdir(str(abs_build_dir)) if os.path.isdir(abs_build_dir / d)]
	# Find all auxiliary modules, they're called "aux1_", "aux2_", etc.,
	# where lower numbers mean fewer dependent levels, so we need to sort
	# them in reversed order to have all dependencies satisfied
	pattern_aux = re.compile(r'^aux\d_.+_llext$')
	aux = sorted([d for d in dirlist if pattern_aux.match(d)], reverse=True)
	llext_dirs = [d for d in dirlist if d.endswith('_llext') and d not in aux]
	final_list = aux + llext_dirs

	for entry in final_list:
		entry_path = abs_build_dir / entry

		uuids = entry_path / 'llext.uuid'
		if not os.path.exists(uuids):
			print(f"Directory {entry} has no llext.uuid file. Skipping.")
			continue

		# replace '_llext' with '.llext', e.g.
		# eq_iir_llext/eq_iir.llext
		llext_base = entry[:-6]
		llext_file = llext_base + '.llext'
		lib_name = ''

		lib_fname = entry_path / 'lib_name.txt'
		if os.path.exists(lib_fname):
			with open(lib_fname, 'r') as libs_f:
				lib_name = libs_f.read()
				if lib_name not in libs.keys():
					libs[lib_name] = []
				libs[lib_name].append(str(entry_path / llext_file))
		else:
			dst = sof_lib_dir / llext_file

			rimage_cfg = entry_path / 'rimage_config.toml'
			llext_input = entry_path / (llext_base + '.llext')
			llext_output = entry_path / (llext_file + '.ri')

			# See why the shlex() parsing step is required at
			# https://docs.zephyrproject.org/latest/develop/west/sign.html#rimage
			# and in Zephyr commit 030b740bd1ec
			sign_cmd = [rimage_cmd, "-o", str(llext_output),
				    "-e", "-c", str(rimage_cfg),
				    "-k", str(signing_key), "-l", "-r"]
			if _ws_args is not None:
				sign_cmd.extend(shlex.split(_ws_args))
			sign_cmd.append(str(llext_input))
			execute_command(sign_cmd, cwd=west_top)

			# An intuitive way to make this multiline would be
			# with (open(dst, 'wb') as fdst, open(llext_output, 'rb') as fllext,
			#	open(llext_output.with_suffix('.llext.xman'), 'rb') as fman):
			# but a Python version, used on Windows errored out on this.
			# Thus we're left with a choice between a 150-character
			# long line and an illogical split like this
			with open(dst, 'wb') as fdst, open(llext_output, 'rb') as fllext, open(
				  llext_output.with_suffix('.ri.xman'), 'rb') as fman:
				# Concatenate the manifest and the llext
				shutil.copyfileobj(fman, fdst)
				shutil.copyfileobj(fllext, fdst)

		# Create symbolic links for all UUIDs
		with open(uuids, 'r') as uuids_f:
			for uuid in uuids_f:
				if os.path.exists(lib_fname):
					if lib_name not in lib_uuids.keys():
						lib_uuids[lib_name] = []
					lib_uuids[lib_name].append(uuid.strip())
				else:
					linkname = uuid.strip() + '.bin'
					symlink_or_copy(sof_lib_dir, llext_file,
							sof_lib_dir, linkname)

	lib_install_dir = sof_output_dir / platform
	if args.key_type_subdir != "none":
		lib_install_dir = lib_install_dir / args.key_type_subdir

	for key in libs.keys():
		lib_path = abs_build_dir / ''.join(['lib', key, '.ri'])
		sign_cmd = [rimage_cmd, "-o", str(lib_path), "-e",
			    "-c", str(abs_build_dir / 'misc' / 'generated' / 'rimage_config_full.toml'),
			    "-k", str(signing_key), "-l", "-r"]
		if _ws_args is not None:
			sign_cmd.extend(shlex.split(_ws_args))
		sign_cmd.extend(libs[key])
		execute_command(sign_cmd, cwd=west_top)
		lib_name = ''.join(['sof-', platform, '-', key, '.ri'])
		dst = lib_install_dir / lib_name
		with open(dst, 'wb') as fdst, open(lib_path, 'rb') as fllext, open(
			  lib_path.with_suffix('.ri.xman'), 'rb') as fman:
			# Concatenate the manifest and the llext
			shutil.copyfileobj(fman, fdst)
			shutil.copyfileobj(fllext, fdst)

			for p_alias in platform_configs[platform].aliases:
				lib_dir = sof_output_dir / p_alias

				if args.key_type_subdir != "none":
					lib_dir = lib_dir / args.key_type_subdir
				lib_dir.mkdir(parents=True, exist_ok=True)
				alias_libname = ''.join(['sof-', p_alias, '-', key, '.ri'])
				symlink_or_copy(lib_install_dir, lib_name,
						lib_dir, alias_libname)

def install_platform(platform, sof_output_dir, platf_build_environ, platform_wconfig):

	# Keep in sync with caller
	platform_build_dir_name = f"build-{platform}"

	# Install to STAGING_DIR
	abs_build_dir = pathlib.Path(west_top) / platform_build_dir_name / "zephyr"

	if args.fw_naming == "AVS":
		# Disguise ourselves for local testing purposes
		output_fwname = "dsp_basefw.bin"
	else:
		# Regular name (deployable build also uses SOF naming convention)
		output_fwname = "".join(["sof-", platform, ".ri"])

	if args.deployable_build and platform_configs[platform].ipc4:
		install_key_dir = pathlib.Path(sof_output_dir, platform)
	else:
		install_key_dir = sof_output_dir

	if args.key_type_subdir != "none":
		install_key_dir = install_key_dir / args.key_type_subdir

	os.makedirs(install_key_dir, exist_ok=True)
	# looses file owner and group - file is commonly accessible
	shutil.copy2(abs_build_dir / "zephyr.ri", install_key_dir / output_fwname)

	if args.deployable_build and platform_configs[platform].ipc4:
		# IPC4 deployable builds are using separate directories per platforms
		# create the structure and symlinks for the alias platforms
		for p_alias in platform_configs[platform].aliases:
			alias_fwname = "".join(["sof-", p_alias, ".ri"])

			alias_key_dir = pathlib.Path(sof_output_dir, p_alias)
			if args.key_type_subdir != "none":
				alias_key_dir = alias_key_dir / args.key_type_subdir

			os.makedirs(alias_key_dir, exist_ok=True)
			symlink_or_copy(install_key_dir, output_fwname, alias_key_dir, alias_fwname)
	else:
		# non deployable builds and IPC3 deployable builds are using the same symlink scheme
		# The production key is usually different
		if args.key_type_subdir != "none" and args.fw_naming != "AVS":
			for p_alias in platform_configs[platform].aliases:
				symlink_or_copy(install_key_dir, output_fwname, install_key_dir, f"sof-{p_alias}.ri")

	if args.deployable_build and platform_configs[platform].ipc4:
		install_lib(platform, sof_output_dir, abs_build_dir, platform_wconfig)


	# sof-info/ directory

	@dataclasses.dataclass
	class InstFile:
		'How to install one file'
		name: pathlib.Path
		renameTo: pathlib.Path = None
		# TODO: upgrade this to 3 states: optional/warning/error
		optional: bool = False
		gzip: bool = True
		txt: bool = False

	installed_files = [
		# Fail if one of these is missing
		InstFile(".config", "config", txt=True),
		InstFile("include/generated/zephyr/autoconf.h", "generated_autoconf.h", txt=True),
		InstFile("include/generated/zephyr/version.h", "zephyr_version.h",
			 gzip=False, txt=True),
		InstFile("include/generated/sof_versions.h", "sof_versions.h",
			 gzip=False, txt=True),
		InstFile(BIN_NAME + ".elf"),
		InstFile(BIN_NAME + ".lst", txt=True, optional=True),
		InstFile(BIN_NAME + ".map", txt=True),

		# Optional because it's not generated for all platforms.
		InstFile(f"misc/generated/rimage_config.toml", "generated_rimage_config.toml",
			 optional=True, gzip=False, txt=True),

		# CONFIG_BUILD_OUTPUT_STRIPPED
		# Renaming ELF files highlights the workaround below that strips the .comment section
		InstFile(BIN_NAME + ".strip", renameTo=f"stripped-{BIN_NAME}.elf"),

		# Not every platform has intermediate rimage modules
		InstFile("main-stripped.mod", renameTo="stripped-main.elf", optional=True),
		InstFile("boot.mod", optional=True),
		InstFile("main.mod", optional=True),
	]

	# We cannot import at the start because zephyr may not be there yet
	sys.path.insert(1, os.path.join(sys.path[0],
					'..', '..', 'zephyr', 'scripts', 'west_commands'))
	import zcmake

	cmake_cache = zcmake.CMakeCache.from_build_dir(abs_build_dir.parent)
	objcopy = cmake_cache.get("CMAKE_OBJCOPY")

	sof_info = pathlib.Path(STAGING_DIR) / "sof-info" / platform
	sof_info.mkdir(parents=True, exist_ok=True)
	gzip_threads = concurrent.ThreadPoolExecutor()
	gzip_futures = []
	for f in installed_files:
		if not pathlib.Path(abs_build_dir / f.name).is_file() and f.optional:
			continue
		dstname = f.renameTo or f.name

		src = abs_build_dir / f.name
		dst = sof_info / dstname

		# Some Xtensa compilers (ab?)use the .ident / .comment
		# section and append the typically absolute and not
		# reproducible /path/to/the.c file after the usual
		# compiler ID.
		# https://sourceware.org/binutils/docs/as/Ident.html
		#
		# --strip-all does not remove the .comment section.
		# Remove it like some gcc test scripts do:
		# https://gcc.gnu.org/git/?p=gcc.git;a=commit;h=c7046906c3ae
		if "strip" in str(dstname):
			execute_command(
				[str(x) for x in [objcopy, "--remove-section", ".comment", src, dst]],
				# Some xtensa installs don't have a
				# XtensaTools/config/default-params symbolic link
				env=platf_build_environ,
			)
		elif f.txt:
			dos2unix(src, dst)
		else:
			shutil.copy2(src, dst)
		if f.gzip:
			gzip_futures.append(gzip_threads.submit(gzip_compress, dst))
	for gzip_res in concurrent.as_completed(gzip_futures):
		gzip_res.result() # throws exception if gzip unexpectedly failed
	gzip_threads.shutdown()

# Zephyr's CONFIG_KERNEL_BIN_NAME default value
BIN_NAME = 'zephyr'

CHECKSUM_WANTED = [
	# Rimage now uses a cryptographic salt and produces a non-deterministic
	# signature; but the ability to match some release with the corresponding
	# build log is still useful.
	'*.ri', '*.llext',
	'dsp_basefw.bin',

	'*version*.h',
	'*autoconf.h', # .config has absolute paths in comments, this has not.
	'*.toml', # rimage
	'*.strip', '*stripped*', # stripped ELF files are reproducible
	'boot.mod', # no debug section -> no need to strip this ELF
	BIN_NAME + '.lst',       # objdump --disassemble
	'*.ldc',
]

# Prefer CRLF->LF because unlike LF->CRLF it's (normally) idempotent.
def dos2unix(in_name, out_name):
	with open(in_name, "rb") as inf:
		# must read all at once otherwise could fall between CR and LF
		content = inf.read()
		assert content
		with open(out_name, "wb") as outf:
			outf.write(content.replace(b"\r\n", b"\n"))

def gzip_compress(fname, gzdst=None):
	gzdst = gzdst or pathlib.Path(f"{fname}.gz")
	with open(fname, 'rb') as inputf:
		# mtime=0 for recursive diff convenience
		with gzip.GzipFile(gzdst, 'wb', mtime=0) as gzf:
			shutil.copyfileobj(inputf, gzf)
	os.remove(fname)


# As of October 2022, sof_ri_info.py expects .ri files to include a CSE manifest / signature.
# Don't run sof_ri_info and ignore silently .ri files that don't have one.
RI_INFO_UNSUPPORTED = []

RI_INFO_UNSUPPORTED += ['imx8', 'imx8x', 'imx8m', 'imx8ulp', 'imx95']
RI_INFO_UNSUPPORTED += ['rn', 'acp_6_0']
RI_INFO_UNSUPPORTED += ['mt8186', 'mt8188', 'mt8195', 'mt8196']

# For temporary workarounds. Unlike _UNSUPPORTED above, the platforms below will print a warning.
RI_INFO_FIXME = [ ]

def reproducible_checksum(platform, ri_file):

	if platform in RI_INFO_FIXME:
		print(f"FIXME: sof_ri_info does not support '{platform}'")
		return

	parsed_ri = sof_ri_info.parse_fw_bin(ri_file, False, False)
	repro_output = ri_file.parent / ("reproducible-" + ri_file.name)
	chk256 = sof_ri_info.EraseVariables(ri_file, parsed_ri, west_top / repro_output)
	print('sha256sum {0}\n{1} {0}'.format(repro_output, chk256))

def create_tarball():
	# vendor directories (typically under /lib/firmware) to include
        # in the tarball
	vendor_dirs = [ 'intel']

	build_top = west_top / 'build-sof-staging' / 'sof'

	tarball_name = 'sof.tgz'

	prev_cwd = os.getcwd()
	os.chdir(build_top)

	tarball_path = build_top / tarball_name
	with tarfile.open(tarball_path, 'w:gz') as tarball:
		for vendor_dir in vendor_dirs:
			vendor_dir_path = build_top / vendor_dir
			if not vendor_dir_path.exists():
				continue
			tarball.add(vendor_dir, recursive=True)
		print(f"Tarball %s created" % tarball_path)

	os.chdir(prev_cwd)

def main():
	parse_args()
	if args.version:
		print(VERSION)
		sys.exit(0)
	check_west_installation()
	if len(args.platforms) == 0:
		print("No platform build requested")
	else:
		print("Building platforms: {}".format(" ".join(args.platforms)))

	west_init_if_needed()

	if args.update:
		# Initialize zephyr project with west
		west_update()
		create_zephyr_sof_symlink()

	build_rimage()
	build_platforms()
	if args.menuconfig:
		return
	if not args.no_tarball:
		create_tarball()
	show_installed_files()

if __name__ == "__main__":
	main()



#!/usr/bin/env python

# SPDX-License-Identifier: ISC
# Copyright (c) 2018, Ulf Magnusson

# Works like 'make olddefconfig', updating an old .config file or creating a
# new one by filing in default values for all new symbols. This is the same as
# picking the default selection for all symbols in oldconfig, or entering the
# menuconfig interface and immediately saving.
#
# The default output filename is '.config'. A different filename can be passed
# in the KCONFIG_CONFIG environment variable.

import os
import sys

import kconfiglib


def main():
    kconf = kconfiglib.standard_kconfig()
    kconf.load_config()
    kconf.write_config()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2023 Intel Corporation. All rights reserved.
#
# The usage of this command is simple. It takes one argument, a path
# to a SOF topology2 source file, and it produces to stdout something
# that can be parsed by Doxygen as C source. Produced output contains
# C struct definitions and instantiations analogous to the topology2
# classes and object instances.
#
# The purpose of the translated C code is not to document actual
# topology2 code, but only to provide anchors for Doxygen to form a
# network of links through which to navigate the topology sources and
# find the pieces of related Doxygen documentation. The filter also
# creates separate pages of the original code and adds links next to the
# pages in the C struct definition, instance documentation and their
# possible Doxygen documentation.

import sys
import os
import re
import io
import logging

logging.basicConfig(filename='filter_debug.txt', filemode='a', encoding='utf-8',
		    level=logging.DEBUG)

def fname():
	try:
		name = sys._getframe(1).f_code.co_name
	except (IndexError, TypeError, AttributeError):
		name = "<unknown>"
	return name

def cbracket_count(line):
	val = line.count("{") - line.count("}")
	return val

def sbracket_count(line):
	val = line.count("[") - line.count("]")
	return val

def doxy_check_add(doxy, line):
	if line.find("##") >= 0:
		doxy = doxy + line[line.find("##"):].replace("##", "//!", 1)
	if line.find("#") >= 0:
		line = line[0:line.find("#")]
	return (doxy, line)

def print_doxy(doxy, file = sys.stdout):
	if len(doxy):
		print(doxy, file=file)
		return ""
	return doxy

def parse_include_str(line):
	if re.search(r"^\s*\<[A-Za-z0-9\/_\-\.]+\>\s*", line):
		tok = line.split()
		return "#include " + tok[0] + "\n"
	return None

def parse_include(line):
	str = parse_include_str(line)
	if str:
		print(str)
		return True
	return False

def parse_define_block(fline, instream):
	"""Parses topology2 Define { } block and outputs C-preprocessor #define macros

	Parameters:
	fline (string): First input line that was read by the caller
	instream (stream): Input stream of topology2 file we are decoding

	Returns:
	string: The original code that was translated
	"""
	if re.search(r"^\s*Define\s+\{", fline):
		logging.debug("fline: %s", fline)
		doxy = ""
		for line in instream:
			if cbracket_count(line) < 0:
				break
			(doxy, line) = doxy_check_add(doxy, line)
			doxy = print_doxy(doxy)
			tok = line.split(maxsplit = 2)
			if len(tok) < 2:
				continue
			val = trim_value(line[line.find(tok[0]) + len(tok[0]):])
			print("#define %s\t%s" % (trim_c_id(tok[0]), val))
		return True
	return False

def parse_include_by_key(fline, instream, file): # For now just skip
	"""Handles IncludeByKey { } blocks, currently handles only actual
	   includes, not nested {} blocks but everything is included into
           raw_code return value.

	Parameters:
	fline (string): First input line that was read by the caller
	instream (stream): Input stream of topology2 file we are decoding
	file (stream): Output stream for the translated output

	Returns:
	string: The original code that was translated

	"""
	if re.search(r"^\s*IncludeByKey\.", fline):
		logging.debug("fline: %s", fline)
		bsum = cbracket_count(fline)
		if bsum == 1:
			# Assume IncludeByKey.<variable> {
			tok = fline.split()
			tok = tok[0].split(".")
			name = tok[1]
			raw_code = ""
			ifstr = "if"
			for line in instream:
				raw_code = raw_code + line
				bsum = bsum + cbracket_count(line)
				tok = line.split("\"")
				if len(tok) >= 4:
					print("#%s %s == \"%s\"\n#include <%s>" %
					      (ifstr, name, tok[1], tok[3]), file = file)
					if ifstr == "if":
						ifstr = "elif"
				if bsum < 1:
					if ifstr != "if":
						print("#endif", file = file)
					return raw_code
	return None

def trim_value(val):
	val = val.strip(" \t\n$")
	end = len(val) - 1
	if val[0:1] == "\"" and val[end:] == "\"":
		return val
	if val.isidentifier() or val.isnumeric():
		return val
	if val[0:1] == "\'" and val[end:] == "\'":
		val = val.strip("\'")
	return "\"" + val + "\""

def trim_c_id(name):
	name = name.strip(" \t\n\"\'")
	name = name.replace("-", "_")
	name = name.replace(" ", "_")
	return name


def parse_attribute_constraints(instream, name):
	"""Parses a Constraints { } block and produces a C enum definition if possible

	Parameters:
	instream (stream): Input stream of topology2 file we are decoding
	name (string): Attribute name of this constraints block belongs to

	Returns:
	string: C enum definition or an empty string

	"""
	logging.debug("name: %s", name)
	valid_values = []
	tuple_values = []
	doxy = "" # This is thrown away since there is no C anchror to connect it-
	raw_code = ""
	enum  = ""
	for line in instream:
		raw_code = raw_code + line
		(doxy, line) = doxy_check_add(doxy, line)
		if cbracket_count(line) < 0:
			break
		if re.search(r"^\s*\!valid_values\s+\[", line):
			for line in instream:
				raw_code = raw_code + line
				(doxy, line) = doxy_check_add(doxy, line)
				if sbracket_count(line) < 0:
					break
				valid_values.append(trim_c_id(line))
		if re.search(r"^\s*\!tuple_values\s+\[", line):
			for line in instream:
				raw_code = raw_code + line
				(doxy, line) = doxy_check_add(doxy, line)
				if sbracket_count(line) < 0:
					break
				tuple_values.append(trim_value(line))
	if len(valid_values) > 1 and len(valid_values) == len(tuple_values):
		enum = "enum " + name + " {\n"
		for i in range(len(valid_values)):
		       enum = enum + "\t\t" + valid_values[i] + " =\t" + tuple_values[i] + ",\n"
		enum = enum + "\t}"
	elif len(valid_values) > 1 and len(tuple_values) == 0:
		enum = "enum " + name + " {\n"
		for i in range(len(valid_values)):
		       enum = enum + "\t\t" + valid_values[i] + ","
		       if valid_values[i][0:1] == "$": # If its a variable add a reference to it
			       enum = enum + " //!< \\ref " + valid_values[i][1:]
		       enum = enum + "\n"
		enum = enum + "\t}"
	return (raw_code, enum)

def parse_class_attribute(attributes, doxy, instream, fline):
	"""Parses a DefineAttribute { } block and collects the information into
	   attributes dict. Any already accumulated Doxygen documentation is
	   also stored there under the attribute name-

	Parameters:
	attributes (dict): A dictionary where data about the attribute is stored
	doxy (strung): Doxygen documentation collected just before the attribute
	instream (stream): Input stream of topology2 file we are decoding
	fline (string): First input line that was read by the caller

	Returns:
	string: The original code that was translated
	"""
	logging.debug("fline: %s", fline)
	tok = fline.split("\"")
	if len(tok) > 1:
		name = tok[1]
	else:
		tok = fline.split(".")
		tok = tok[1].split(" ")
		name = tok[0]
	(doxy, fline) = doxy_check_add(doxy, fline)
	bsum = cbracket_count(fline)
	typestr = ""
	token_ref = ""
	ref_type = ""
	enum = ""
	raw_code = ""
	if bsum < 1:
		# Assume: DefineAttribute.name {}
		typestr = "int"
	else:
		for line in instream:
			raw_code = raw_code + line
			(doxy, line) = doxy_check_add(doxy, line)
			bsum = bsum + cbracket_count(line)
			if bsum < 1:
				break
			if re.search(r"^\s*constraints\s+\{", line) and bsum == 2:
				(code, enum) = parse_attribute_constraints(instream, name)
				raw_code = raw_code + code
				bsum = 1
			elif re.search(r"^\s*type\s", line):
				tok = line.split()
				typestr = tok[1].strip(" \t\n\"\'")
			elif re.search(r"^\s*token_ref\s", line):
				tok = line.split()
				token_ref = tok[1].strip(" \t\n\"\'")
				tok = token_ref.split(".")
				ref_type = tok[1]
	logging.debug("type %s token_ref %s ref_type %s enum %s",
		      typestr, token_ref, ref_type, enum)
	if enum != "" and ref_type != "bool":
		typestr = enum
		if ref_type == "string":
			doxy = doxy + "//! \\em string type\n"
	elif typestr == "" or ref_type == "bool":
		typestr = ref_type
	attributes[name] = { "type": typestr, "doxy": doxy, "token_ref": token_ref }
	return raw_code

def print_attributes(attributes):
	"""Print out struct members and their doxygen documentation from attributes dict.

	Parameters:
	attributes (dict): A dictionary where data about the attributes was stored
	"""
	for name in attributes:
		doxy = ""
		if attributes[name].get("doxy"):
			doxy = attributes[name]["doxy"] + "\n"
		if attributes[name].get("type"):
			typestr = attributes[name]["type"]
			print("%s\t%s %s;\n" % (doxy, typestr, name))

def set_attribute_flag(attributes, attribute, flag):
	if not attributes.get(attribute):
		attributes[attribute] = {}
	attributes[attribute][flag] = True

def parse_attributes_block(instream, attributes, attrib_doxy):
	"""Parse attributes block inside class definition and store the flags in attributes dict

	Parameters:
	instream (stream): Input stream of topology2 file we are decoding
	attributes (dict): Dict where data about the attributes is stored
	attrib_doxy (dict): Dict to store doxygen docs associated with the attributes block

	Returns:
	string: The original code that was translated
	"""
	logging.debug("called")
	raw_code = ""
	doxy = ""
	for line in instream:
		raw_code = raw_code + line
		(doxy, line) = doxy_check_add(doxy, line)
		if cbracket_count(line) < 0:
			break
		if re.search(r"^\s*\!constructor\s+\[", line):
			for line in instream:
				raw_code = raw_code + line
				(doxy, line) = doxy_check_add(doxy, line)
				if sbracket_count(line) < 0:
					break
				set_attribute_flag(attributes, trim_c_id(line), "constructor")
			attrib_doxy["constructor"] = doxy
			doxy = ""
		elif re.search(r"^\s*\!mandatory\s+\[", line):
			for line in instream:
				raw_code = raw_code + line
				(doxy, line) = doxy_check_add(doxy, line)
				if sbracket_count(line) < 0:
					break
				set_attribute_flag(attributes, trim_c_id(line), "mandatory")
			attrib_doxy["mandatory"] = doxy
			doxy = ""
		elif re.search(r"^\s*\!immutable\s+\[", line):
			for line in instream:
				raw_code = raw_code + line
				(doxy, line) = doxy_check_add(doxy, line)
				if sbracket_count(line) < 0:
					break
				set_attribute_flag(attributes, trim_c_id(line), "immutable")
			attrib_doxy["immutable"] = doxy
			doxy = ""
		elif re.search(r"^\s*\!deprecated\s+\[", line):
			attrib_doxy["deprecated"] = doxy
			doxy = ""
			for line in instream:
				raw_code = raw_code + line
				(doxy, line) = doxy_check_add(doxy, line)
				if sbracket_count(line) < 0:
					break
				set_attribute_flag(attributes, trim_c_id(line), "deprecated")
		elif re.search(r"^\s*unique\s+", line):
			tok = line.split()
			attrib_doxy["unique"] = doxy
			doxy = ""
			set_attribute_flag(attributes, trim_c_id(tok[1]), "unique")
	return raw_code

def attribute_block_print(class_name, attributes, attrib_doxy):
	"""Generates a Doxygen documentation section from attribute flags
	   and doxygen docs stored to attrib_doxy

	Parameters:
	class_name (string): Name of the class we are processing
	attributes (dict): Dict where data about the attributes was stored
	attrib_doxy (dict): Dict where doxygen docs of the attributes block was stored

	"""
	logging.debug("class_name \'%s\'", class_name)
	for field in attrib_doxy:
		# Only add attribute paragraph if there are doxygen comment for it
		if attrib_doxy[field] != "":
			print("//! \\par %s attributes:\n//!" % field.capitalize())
			for attr in attributes:
				if attributes[attr].get(field):
					print("//! \\link %s::%s \\endlink \\n" % (class_name, attr))
			print("//! \\n\n%s" % attrib_doxy[field])

def attribute_block_info_add(attributes, attrib_doxy):
	"""Add simple bullets in the attribute (= struct members) documentation if
	   the attribute has constructor, mandatory, immutable, deprecated, or unique
	   property.

	Parameters:
	attributes (dict): Dict where data about the attributes was stored
	attrib_doxy (dict): Dict where doxygen docs of the attributes block was stored

	"""
	# TODO: Add short documentation about the attribute properties and link to it.
	for field in attrib_doxy:
		for attr in attributes:
			if attributes[attr].get(field):
				if not attributes[attr].get("doxy"):
					attributes[attr]["doxy"] = ""
				attributes[attr]["doxy"] = "//! - \\em " + field + " attribute.\\n\n" + attributes[attr]["doxy"]

def parse_class_contents(class_name, attributes, attrib_doxy, objs, defaults,
			 instream, includef = ""):
	"""Translates the contents inside a class definition { } block

	Parameters:
	class_name (string): Class name
	attributes (dict): Dict where data about the attributes is stored
	attrib_doxy (dict): Dict to store doxygen docs associated with the attributes block
	objs (dict): Dict to store fist level contained objects with Docxyge docs
	defaults (streaM): Stream where the instances of the objects are printed
	instream (stream): Input stream of topology2 file we are decoding
	includef (string): Include file from where the attributes and objects inline included from

	Returns:
	string: The original code that was translated
	"""
	bsum = 1
	doxy_addition = ""
	if includef != "":
		doxy_addition = "//! - Included from <" + includef + ">\\n\n"
	doxy = ""
	raw_code = ""
	for line in instream:
		raw_code = raw_code + line
		if parse_include_str(line):
			# Inline files that are included from within a class definition
			filename = line[line.find("<") + 1:line.find(">")]
			logging.debug("try to inline include \'%s\' from \'%s\'",
				      filename, os.getcwd())
			# NOTE: The path is relative to when the script exists so
			#       two levels up and we are at topology2 root
			with open("../../" + filename, "r+", encoding="ascii") as ifile:
				parse_class_contents(class_name, attributes, attrib_doxy, objs,
						     defaults, ifile, filename)
			continue
		if (code := parse_include_by_key(line, instream, sys.stdout)):
			raw_code = raw_code + code
			continue
		if re.search(r"^\s*DefineAttribute\.", line):
			doxy = doxy_addition + doxy
			raw_code = raw_code + parse_class_attribute(attributes, doxy, instream, line)
			doxy = ""
		elif re.search(r"^\s*DefineArgument\.", line):
			for line in instream: # Just skip, this is only used in bytes.conf
				if cbracket_count(line) < 0:
					break
		elif re.search(r"^\s*attributes\s+\{", line):
			doxy = "" # Doxy comments before attributes block end in weird places
			raw_code = raw_code + parse_attributes_block(instream, attributes, attrib_doxy)
		elif re.search(r"^\s*Object\.", line):
			# TODO: Pass collected doxy comments to parse_object and store in objs
			doxy = ""
			raw_code = raw_code + parse_object(instream, line, file = defaults,
							   objects = objs, tabs = "\t", ending = ",")
		else: # If nothing else matched, assume a default value definition for an attribute
			(doxy, line) = doxy_check_add(doxy, line)
			tok = line.split()
			if len(tok) == 2 and tok[0].isidentifier():
				if doxy != "":
					print(doxy, file = defaults)
					doxy = ""
				val = trim_value(line[line.find(tok[0]) + len(tok[0]):])
				print("\t.%s =\t%s," % (trim_c_id(tok[0]), val),
				      file = defaults)
			bsum = bsum + cbracket_count(line)
			if bsum < 1:
				break
	return raw_code

def parse_class(instream, fline):
	"""Parse class definition and print out C struct definition

	Parameters:
	fline (string): First input line that was read by the caller
	instream (stream): Input stream of topology2 file we are decoding
	"""
	logging.debug("fline: %s", fline)
	tok = fline.split(maxsplit = 2)
	cdef = tok[0].split(".")
	# base = cdef[1] # Just in case we go back to C++ tralation
	class_name = cdef[2]
	class_name = trim_c_id(class_name)
	attributes = {}
	attrib_doxy = {}
	defaults = io.StringIO()
	objs = {}
	raw_code = fline + parse_class_contents(class_name, attributes, attrib_doxy,
						objs, defaults, instream)
	attribute_block_print(class_name, attributes, attrib_doxy)
	attribute_block_info_add(attributes, attrib_doxy)
	print("//! \\ref %s_rawcode" % class_name)
	print("struct %s {" % class_name)
	print_attributes(attributes)
	for obj in objs:
		if objs[obj]["count"] > 1:
			for i in range(objs[obj]["count"]):
				print(objs[obj]["doxy"][i])
				print("\tstruct %s %s%d;" % (obj, obj, i))
		else:
			print(objs[obj]["doxy"][0])
			print("\tstruct %s %s;" % (obj, obj))
	print("};\n")
	print("/*! \\page %s_rawcode The %s class definition in topology2 code\n\t\\code{.unparsed}" %
	      (class_name, class_name))
	print(raw_code)
	print("\t\\endcode\n*/")
	print("//! \\var struct %s %s_defaults" % (class_name, class_name))
	print("//! \\brief %s class default values" % class_name)
	print("struct %s %s_defaults = {\n" % (class_name, class_name))
	print(defaults.getvalue())
	print("};\n")
	defaults.close()

def parse_members(instream, cbsum, tabs, file):
	"""Parse attribute initializations from object instantiation

	Parameters:
	instream (stream): Input stream of topology2 file we are decoding
	cbsum (int): The amount of curly brackets "{" WE HAVE OPEN
	tabs (string): Current level of indentation

	Returns:
	string: The original code that was translated
	"""
	doxy = ""
	raw_code = ""
	for line in instream:
		raw_code = raw_code + line
		if (code := parse_include_by_key(line, instream, file)):
			raw_code = raw_code + code
			continue
		cbsum = cbsum + cbracket_count(line)
		(doxy, line) = doxy_check_add(doxy, line)
		# Assume ending } to be alone on its own line
		if cbsum < 1:
			break
		if re.search(r"^\s*Object\.", line):
			logging.debug("object-line: %s", line)
			obj = line.split(".")
			if cbsum == 2:
				# Assume Object.Base.name.1 {
				doxy = print_doxy(doxy, file = file)
				raw_code = raw_code + parse_object(instream, line, file = file, tabs = tabs, ending = ",")
				cbsum = 1
			elif cbsum == 1:
				# Assume  Object.Base.name.1 {}
				doxy = print_doxy(doxy, file = file)
				print("%s.%s = {}," % (tabs, obj[2]), file = file)
		else:
			tok = line.split(maxsplit = 1)
			if len(tok) >=2:
				name = tok[0]
				val = trim_value(tok[1])
				doxy = print_doxy(doxy, file = file)
				print("%s.%s = %s," % (tabs, name, val), file = file)
	return raw_code

#
def object_instance_prefix(name, ending):
	"""Decide "struct name name =" or ".name =" based on instantiation ending in ',' or ';'

	Parameters:
	name (string): Name of the object instance
	ending (string): Either ',' or ';' indication if this is an instance or a definition

	"""
	if ending == ";":
		return "struct " + name + " "
	return "."

def parse_object(instream, fline, file = sys.stdout, objects = {}, tabs = "", ending = ";"):
	"""Translates all Object instatiation into initialized C structs
	   Note that dict arguments in python are passed as reference

	Parameters:
	instream (stream): Input stream of topology2 file we are decoding
	fline (string): First input line that was read by the caller
	file (stream): Where the C struct instance or definition is printed
	objects (dict): Dict to store fist level contained objects with Docxyge docs
	ending (string): Either ',' or ';' indication if this is an instance or a definition
	tabs (string): Current level of indentation

	Returns:
	string: The original code that was translated
	"""
	logging.debug("fline: %s", fline)
	tok = fline.split(maxsplit = 2)
	obj = tok[0].split(".")
	cbsum = cbracket_count(fline)
	sbsum = sbracket_count(fline)
	name = ""
	doxy = ""
	raw_code = ""
	if len(obj) == 4 and sbsum == 0 and cbsum == 0:
		# Assume Object.Base.name.1 { }
		name = obj[2]
		name = trim_c_id(name)
		doxy = print_doxy(doxy, file = file)
		prefix = object_instance_prefix(name, ending)
		print("%s%s = {}%s" % (prefix, name, ending), file = file)
		objects[name] = { "count": 1, "doxy": [doxy] }
	elif len(obj) == 4 and sbsum == 0 and cbsum == 1:
		# Assume Object.Base.name.1 {
		name = obj[2]
		name = trim_c_id(name)
		doxy = print_doxy(doxy, file = file)
		prefix = object_instance_prefix(name, ending)
		print("%s%s%s = {" % (tabs, prefix, name), file = file)
		raw_code = raw_code + parse_members(instream, cbsum, tabs + "\t", file)
		print("%s}%s" % (tabs, ending), file = file)
		objects[name] = { "count": 1, "doxy": [doxy] }
	elif len(obj) == 3 and sbsum == 1 and cbsum == 0:
		# Assume Object.Base.name [
		name = obj[2]
		name = trim_c_id(name)
		doxy = print_doxy(doxy, file = file)
		prefix = object_instance_prefix(name, ending)
		print("%s%s%s[] = {" % (tabs, prefix, name), file = file)
		objects[name] = { "count": 0, "doxy": [] }
		for line in instream:
			raw_code = raw_code + line
			if parse_include(line):
				continue
			if (code := parse_include_by_key(line, instream, sys.stdout)):
				raw_code = raw_code + code
				continue
			sbsum = sbsum + sbracket_count(line)
			cbsum = cbsum + cbracket_count(line)
			(doxy, line) = doxy_check_add(doxy, line)
			if sbsum < 1:
				print("%s}%s" % (tabs, ending), file = file)
				break
			if cbsum == 1: # Assume starting { on its own line
				objects[name]["count"] = objects[name]["count"] + 1
				objects[name]["doxy"].append(doxy)
				doxy = print_doxy(doxy, file = file)
				print("%s\t{" % tabs, file = file)
				raw_code = raw_code + parse_members(instream, cbsum, tabs + "\t\t", file)
				cbsum = 0
				print("%s\t}," % tabs, file = file)
	elif len(obj) == 2 and sbsum == 0 and cbsum == 1:
		# Assume Object.Base {
		for line in instream:
			raw_code = raw_code + line
			if parse_include(line):
				continue
			if (code := parse_include_by_key(line, instream, file)):
				raw_code = raw_code + code
				continue
			sbsum = sbsum + sbracket_count(line)
			cbsum = cbsum + cbracket_count(line)
			(doxy, line) = doxy_check_add(doxy, line)
			if cbsum < 1: # Ending } found
				break
			if sbsum == 1 and cbsum == 1 and sbracket_count(line) > 0:
				# Assume Class_name [
				tok = line.split()
				name = trim_c_id(tok[0])
				objects[name] = { "count": 0, "doxy": [] }
				doxy = print_doxy(doxy, file = file)
				prefix = object_instance_prefix(name, ending)
				print("%s%s%s[] = {" % (tabs, prefix, name), file = file)
			elif sbsum == 1 and cbsum == 2:
				# Assume class_name [ \n { \n
				objects[name]["count"] = objects[name]["count"] + 1
				objects[name]["doxy"].append(doxy)
				doxy = print_doxy(doxy, file = file)
				print("%s\t{" % tabs, file = file)
				raw_code = raw_code + parse_members(instream, 1, tabs + "\t\t", file)
				print("%s\t}," % tabs, file = file)
				cbsum = 1
			elif sbsum == 0 and cbsum == 1 and sbracket_count(line) < 0:
				# Assume ending ] of class table alone one its own line
				print("%s}%s" % (tabs, ending), file = file)
			elif sbsum == 0 and cbsum == 1 and line.count("}") == 1:
				# Assume name."1" {}
				# No init values, so no instantiation code needed, but we still need
				# to store the object into objexts dict.
				tok = line.split()
				tok = tok[0].split(".")
				name = trim_c_id(tok[0])
				if not objects.get(name):
					objects[name] = { "count": 0, "doxy": [] }
				objects[name]["count"] = objects[name]["count"] + 1
				objects[name]["doxy"].append(doxy)
				doxy = print_doxy(doxy, file = file)
			elif sbsum == 0 and cbsum == 2 and cbracket_count(line) == 1:
				# Assume name."1" {
				tok = line.split()
				tok = tok[0].split(".")
				name = trim_c_id(tok[0])
				mname = name
				if len(tok) > 1:
					idx = tok[1].strip(" \"\'")
					if idx.isidentifier() or idx.isnumeric():
						mname = mname + "_" + idx
				if not objects.get(name):
					objects[name] = { "count": 0, "doxy": [] }
				objects[name]["count"] = objects[name]["count"] + 1
				objects[name]["doxy"].append(doxy)
				doxy = print_doxy(doxy, file = file)
				prefix = object_instance_prefix(name, ending)
				print("%s%s%s = {" % (tabs, prefix, mname), file = file)
				raw_code = raw_code + parse_members(instream, 1, tabs + "\t", file)
				cbsum = 1
				print("%s}%s" % (tabs, ending), file = file)
	return raw_code

def parse_object_and_make_raw_code_page(filename, index, instream, fline):
	"""Parses an object instance outputs its C equivalent, and creates a raw code page of
	   of it and prints a reference to it. Handles an "Object.... {}" instance completely
	   and produces possibly multiple initialized C structs, but always just one raw code
	   block containing the "Object.... {}" block completely.

	Parameters:
	filename (stream): Name of the file we are paring
	index (int): The index of the decoded Object block in this file we are handling
	instream (stream): Input stream of topology2 file we are decoding
	fline (string): First input line that was read by the caller

	"""
	tok = filename.split("/")
	filename = fname = tok[len(tok)-1]
	fname = fname.replace("-", "_")
	fname = fname.replace(".", "_")
	c_instances = io.StringIO()
	raw_code = fline + parse_object(instream, fline, c_instances)
	print("/*! \\page %s_%d_rawcode The %s instances #%d in topology2 code\n\t\\code{.unparsed}" %
	      (fname, index, filename, index))
	print(raw_code)
	print("\t\\endcode\n*/")
	print("//! \\brief \\ref %s_%d_rawcode" % (fname, index))
	print(c_instances.getvalue())
	c_instances.close()

# Main starts here, apart from debug file opening
filename = sys.argv[1]
logging.info("file: %s", filename)
with open(filename, "r+", encoding="ascii") as instream:
	block_idx = 0

	shortfname = filename[filename.find("/topology2/"):]
	print("//! \\file %s" % shortfname[11:])
	print("//! Source file can be found " +
	      "<a href=\"https://github.com/thesofproject/sof/tree/main/tools/topology%s\">here</a>."
	      % shortfname)
	for line in instream:
		if parse_include(line):
			continue
		if parse_define_block(line, instream):
			continue
		if parse_include_by_key(line, instream, sys.stdout):
			continue
		if re.search(r"^\s*Class\.", line):
			parse_class(instream, line)
		elif re.search(r"^\s*Object\.", line):
			parse_object_and_make_raw_code_page(filename, block_idx, instream, line)
			block_idx = block_idx + 1
		elif line.find("##") >= 0:
			sys.stdout.write(line.replace("##", "//!", 1))
		else:
			sys.stdout.write("\n")
